/*
  Name:        	Do178Builder-utility.c
  Author:      	Ron Burkey
  Purpose:     	This is a command-line utility that duplicates some of the
  		processing capabilities of Do178Builder.  In other words, 
		it reads Do178Builder-compatible XML files, and then 
		produces certain types of output from it.  This functionality
		could be incorporated into Do178Builder itself, but the
		overhead of maintaining a GUI just to get some processing
		seems silly.
  Mod History:	03/13/2008 RSB	Began.
  		03/17/2008 RSB	Added some --trace-verify stuff.  Replaced
				--regex-extended by --regex-basic.
		03/18/2008 RSB	--trace-system related stuff.  Changed
				default --high-section-pattern and 
				--low-section-pattern so that punctuation
				could immediately precede or follow the
				"shall".
  
  Copyright 2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/

// IS_DO178BUILDER_UTILITY_MAIN is only in this file, and must be #defined
// prior to #including Do178Builder-utility.h.  Its purpose is to cause
// global variables to be allocated here, and extern'd in all other source
// files.
#define IS_DO178BUILDER_UTILITY_MAIN
#include "Do178Builder-utility.h"

// A simple macro for getting a command-line option of the form "--something".
#define ELSE_SWITCH_SIMPLE(Switch,Field) \
  else if (!strcmp (argv[i], Switch)) \
    { \
      Options.Field = 1; \
    }
// A macro for getting a command-line option of the form "--something=string".
#define ELSE_SWITCH_STRING(Switch,Field) \
  else if (!strncmp (argv[i], Switch, strlen (Switch))) \
    { \
      Options.Field = &argv[i][strlen (Switch)]; \
    }
// A macro for getting a command-line option of the form "--something=char".
#define ELSE_SWITCH_CHAR(Switch,Field) \
  else if (!strncmp (argv[i], Switch, strlen (Switch))) \
    { \
      Options.Field = argv[i][strlen (Switch)]; \
    }

// 'sed' commands for transforming Do178Builder database XML (which is 
// illegally formed in some tag names.  There are three issues, which 
// are addressed by the commands SED1, SED2, and SED3.
// Issue #1 is entities like &#NN; when they appear in the
// tag names.  Since the characters '&', '#', and ';' cannot appear legally
// in tag names, we need to remove them.  As a purely arbitrary measure, 
// we use 'sed' to replace "&#NN;" with "_rNNr_".  The actual sed command
// we're trying to achieve is something like this:
//	sed "s/&#\([0-9][0-9]\);/_r\1r_/g" 
// Actually, we translate throughout the document, not just in tag names,
// since there's no reason not to.
#define SED1 "\"s/&#\\([0-9][0-9]\\);/_r\\1r_/g\""
// Issue #2 is correction of tag names beginning with a digit.
// The correction is to envelope the leading digit _ _:
//	sed "s/<\([0-9]\)/<_\1_/g"
#define SED2 "\"s/<\\([0-9]\\)/<_\\1_/g\""
// Issue #3 is correction of closing tag names beginning
// with a digit.
//	sed "s/<\/\([0-9]\)/<\/_\1_/g"
#define SED3 "\"s/<\\/\\([0-9]\\)/<\\/_\\1_/g\""
      
// A dummy string.  Global merely to avoid overburdening the stack.
static char Command[1024];


// Compile a regular expression.  Return 0 on success, non-zero on failure.
int
Pattern (regex_t *Machine, const char *PatternString, ProcessingOptions_t *Options)
{    
  if (PatternString != NULL)
    {
      int RegexOptions;
      
      RegexOptions = REG_NEWLINE;
      if (Options->RegexCaseInsensitive)
        RegexOptions |= REG_ICASE;
      if (Options->RegexExtended)
        RegexOptions |= REG_EXTENDED;
      
      if (regcomp (Machine, PatternString, RegexOptions))
        {
	  fprintf (stderr, 
	           "Compilation of regular expression \"%s\" failed\n", 
	           PatternString);
	  return (1);
	}
    } 
  return (0);
}   
    
// A function, suitable for use with WalkNodes(), that can display the name 
// of a node in the XML tree.

int 
PrintNodeName (xmlNode *a_node, ProcessingOptions_t *Options)
{
  // Print the tag name of the section.
  printf ("%*s%s", 5 * (Options->WalkLevel - 1), "", a_node->name);
  
  // Print the documents this section is a part of.
  if (Options->CombineSrdSdd)
    {
      if (Options->DocSections[REQ_DOC].Present)
	printf (" (%s %s)", Options->ComDocName, Options->DocSections[REQ_DOC].Number);
      if (Options->DocSections[VER_DOC].Present)
	printf (" (%s %s)", Options->VerDocName, Options->DocSections[VER_DOC].Number);
    }
  else
    {
      if (Options->DocSections[REQ_DOC].Present)
	printf (" (%s %s)", Options->ReqDocName, Options->DocSections[REQ_DOC].Number);
      if (Options->DocSections[DES_DOC].Present)
	printf (" (%s %s)", Options->DesDocName, Options->DocSections[DES_DOC].Number);
      if (Options->DocSections[VER_DOC].Present)
	printf (" (%s %s)", Options->VerDocName, Options->DocSections[VER_DOC].Number);
    }
  printf ("\n");
  return (0);
}

///////////////////////////////////////////////////////////////////////////////

int
main (int argc, char *argv[])
{
  ProcessingOptions_t Options = { NULL };
  int i, j, RetVal = 1;
  char *TempFilename = "Do178Builder-utility.xml";
  xmlDocPtr doc; 	// The in-memory document tree

  fprintf (stderr, "Do178Builder-utility, build " __DATE__ ", " __TIME__ "\n");
  fprintf (stderr, "(c)" COPYRIGHT " by Ronald S. Burkey <info@sandroid.org>, wwww.do-178b.org\n");

  //----------------------------------------------------------------------------
  // Process the command-line arguments.
  Options.DelimiterImplement = ',';
  Options.DelimiterSystem = ',';
  Options.ReqDocName = "SRD";
  Options.DesDocName = "DD";
  Options.VerDocName = "SVCP";
  Options.ComDocName = "SRD/SDD";
  Options.HighSectionPattern = "[[:space:][:punct:]]shall[[:space:][:punct:]]";
  Options.LowSectionPattern = "[[:space:][:punct:]]shall[[:space:][:punct:]]";
  Options.HighLabelPattern = "\\[SR[0-9]+\\]";
  Options.LowLabelPattern = "\\[SL[0-9]+\\]";
  Options.RegexExtended = 1;
  
  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] == '-')	// OPTION?
        {
	  if (!strcmp (argv[i], "--help"))
	    {
	      printf ("USAGE:\n");
	      printf ("\tDo178Builder-utility [OPTIONS] DocumentDatabase.xml\n");
	      printf ("The allowed options are:\n");
	      printf ("--combine-srd-sdd        Assume that the SRD and SDD (or the HRD\n");
	      printf ("                         and HDRD) are combined into a single\n");
	      printf ("                         document.\n");
	      printf ("--delimiter-implement=C  For <trace></trace> blocks that define\n");
	      printf ("                         where in the source code or CAD files a \n");
	      printf ("                         given low-level requirement is implemented,\n");
	      printf ("                         defines the delimiter that separates\n");
	      printf ("                         the requirements.  Usually a space or a\n");
	      printf ("                         comma.  If not a space, then leading or\n");
	      printf ("                         trailing spaces around the delimiter will\n");
	      printf ("                         be trimmed.  In either case, multiple spaces\n");
	      printf ("                         will be collapsed to a single space.\n");
	      printf ("                         Defaults to a space.\n");
	      printf ("--delimiter-system=C     Same as --delimiter-implement, but for \n");
	      printf ("                         traceability of system requirements to \n");
	      printf ("                         high-level requirements.\n");
	      printf ("--display                Simply display what was read in.\n");
	      printf ("--DO-254                 Look for HRD/HDRD/HTPR documents instead\n");
	      printf ("                         the default SRD/SDD/SVCP documents.\n");
	      printf ("--help                   Display this message.\n");           
	      printf ("--high-label-pattern=P   Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-XXXX switches to\n");
	      printf ("                         identify labels for high-level software\n");
	      printf ("                         (hardware) requirements in the SRD (HRD).\n");
	      printf ("--high-section-pattern=P Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-XXXX switches to\n");
	      printf ("                         identify sections  in the SRD (HRD)\n");
	      printf ("                         containing unique high-level software\n");
	      printf ("                         (hardware) requirements identified by\n");
	      printf ("                         section number.\n");
	      printf ("--high-section-range=P   Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-XXXX switches to\n");
	      printf ("                         constrain the range of section numbers\n");
	      printf ("                         in which high-level software (hardware)\n");
	      printf ("                         requirements are sought.\n");
	      printf ("--low-label-pattern=P    Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-XXXX switches to\n");
	      printf ("                         identify labels for low-level software\n");
	      printf ("                         (hardware) requirements in the SDD (HDRD).\n");
	      printf ("--low-section-pattern=P  Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-XXXX switches to\n");
	      printf ("                         identify sections  in the SDD (HDRD)\n");
	      printf ("                         containing unique low-level software\n");
	      printf ("                         (hardware) requirements identified by\n");
	      printf ("                         section number.\n");
	      printf ("--low-section-range=P    Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-XXXX switches to\n");
	      printf ("                         constrain the range of section numbers\n");
	      printf ("                         in which low-level software (hardware)\n");
	      printf ("                         requirements are sought.\n");
	      printf ("--no-sed                 Do NOT run \'sed\' to translate the\n");
	      printf ("                         entity names in the input file.\n");
	      printf ("--regex-basic            Use POSIX Basic Regular Expression\n");
	      printf ("                         syntax rather than POSIX Extended Regular\n");
	      printf ("                         Expression syntax, which is the default.\n");
	      printf ("--regex-case-insensitive Make regular-expression matching \n");
	      printf ("                         insensitive to upper/lower case\n");
	      printf ("                         distinction.\n");
	      printf ("--sortable=N             Stringify document section numbers in a\n");
	      printf ("                         way that's more easily sortable within\n");
	      printf ("                         spreadsheets.  For example, I don\'t like it\n");
	      printf ("                         that a spreadsheet will think that 10.1\n");
	      printf ("                         precedes 2.3.4.5.  The technique is to \n");
	      printf ("                         left-pad the numerical fields with zeroes.\n");
	      printf ("                         N is the field size.  N=2 would be the most\n");
	      printf ("                         common setting, but I can imagine a need for\n");
	      printf ("                         the occasional N=3.\n");
	      printf ("--system-section-pattern=P Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-system switch to\n");
	      printf ("                         identify sections  in the SRD (HRD)\n");
	      printf ("                         containing cross-references to system\n");
	      printf ("                         requirements external to the SRD.  The\n");
	      printf ("                         cross-references are of the form:\n");
	      printf ("                              <trace>list of reqs.</trace>\n");
	      printf ("--trace                  Activate all of the --trace-XXXX options\n");
	      printf ("                         listed below.\n");
	      printf ("--trace-design           Output traceability data between\n");
	      printf ("                         SRD and SDD (or HRD and HDRD) into\n");
	      printf ("                         the file Do178Builder-trace-design.csv.\n");
	      printf ("--trace-design-full      Same as --trace-design, but includes\n");
	      printf ("                         SRD/SDD (HRD/HDRD) sections without\n");
	      printf ("                         software (hardware) requirements.\n");
	      printf ("--trace-system           Output traceability data between\n");
	      printf ("                         system requirements and SRD, into the\n");
	      printf ("                         file Do178Builder-trace-system.csv.\n");
	      printf ("--trace-implement        Output traceability data between SDD\n");
	      printf ("                         (or HDRD) and source code (or CAD),\n");
	      printf ("                         into Do178Builder-trace-implement.\n");
	      printf ("--trace-verify           Output traceability data between\n");
	      printf ("                         SRD/SDD (or HRD/HDRD) and SVCP (or\n");
	      printf ("                         HTPR), in Do178Builder-trace-verify.csv.\n");
	      printf ("--verify-section-range=P Defines a pattern (regular expression)\n");
	      printf ("                         used with the --trace-verify switches to\n");
	      printf ("                         constrain the range of section numbers\n");
	      printf ("                         in which test cases are sought.\n");
	      goto Done;
	    }
	  else if (!strcmp (argv[i], "--DO-254"))
	    {
	      Options.ReqDocName = "HRD";
	      Options.DesDocName = "HDRD";
	      Options.VerDocName = "HTPR";
	      Options.ComDocName = "HRD/HDRD";
	    }
	  else if (!strcmp (argv[i], "--trace"))
	    {
	      Options.TraceDesign = 1;
	      Options.TraceDesignFull = 1;
	      Options.TraceSystem = 1;
	      Options.TraceImplement = 1;
	      Options.TraceVerify = 1;
	    }
	  else if (1 == sscanf (argv[i], "--sortable=%d", &j))
  	    {
              if (j != 2 && j != 3)
                {
                  fprintf (stderr, "Error: For --sortable=N, N must be 2 or 3.\n");
                  goto Done;
                }
	      Options.Sortable = j;
            }
	    
	  // Parse all of the simpler command-line switches;
	  // (I know that the empty comments at the ends of the lines may be 
	  // confusing.  They're there to prevent unwanted reformatting when
	  // I run this source code through the 'indent' utility.)  
	  ELSE_SWITCH_SIMPLE ("--combine-srd-sdd", CombineSrdSdd) //
	  ELSE_SWITCH_CHAR ("--delimiter-implement=", DelimiterImplement) //
	  ELSE_SWITCH_CHAR ("--delimiter-system=", DelimiterSystem) //
	  ELSE_SWITCH_SIMPLE ("--display", Display) //
	  ELSE_SWITCH_STRING ("--high-label-pattern=", HighLabelPattern) //
	  ELSE_SWITCH_STRING ("--high-section-pattern=", HighSectionPattern) //
	  ELSE_SWITCH_STRING ("--high-section-range=", HighSectionRange) //
	  ELSE_SWITCH_STRING ("--low-label-pattern=", LowLabelPattern) //
	  ELSE_SWITCH_STRING ("--low-section-pattern=", LowSectionPattern) //
	  ELSE_SWITCH_STRING ("--low-section-range=", LowSectionRange) //
	  ELSE_SWITCH_SIMPLE ("--no-sed", NoSed) //
	  ELSE_SWITCH_SIMPLE ("--regex-case-insensitive", RegexCaseInsensitive) //
	  ELSE_SWITCH_SIMPLE ("--regex-basic", RegexBasic) //
	  ELSE_SWITCH_STRING ("--system-section-range=", SystemSectionRange) //
	  ELSE_SWITCH_SIMPLE ("--trace-design", TraceDesign) //
	  ELSE_SWITCH_SIMPLE ("--trace-design-full", TraceDesignFull) //
	  ELSE_SWITCH_SIMPLE ("--trace-system", TraceSystem) //
	  ELSE_SWITCH_SIMPLE ("--trace-implement", TraceImplement) //
	  ELSE_SWITCH_SIMPLE ("--trace-verify", TraceVerify) //
	  ELSE_SWITCH_STRING ("--verify-section-range=", VerifySectionRange) //
	  
	  // Nope, didn't recognize this switch.
	  else
	    {
	      fprintf (stderr, "Error: Unknown command-line switch \"%s\"\n", argv[i]);
	      goto Done;
	    }
	}
      else if (Options.DatabaseName != NULL)
        {
	  fprintf (stderr, "Error: Database name defined twice (\"%s\" and \"%s\")\n",
	  	  Options.DatabaseName, argv[i]);
	  goto Done;
	}
      else
        {
	  Options.DatabaseName = argv[i];
	}
    }
  if (Options.DatabaseName == NULL)
    {
      fprintf (stderr, "Error: No database name was specified.\n");
      goto Done;
    }
  if (Options.RegexBasic)
    Options.RegexExtended = 0;
    
  //----------------------------------------------------------------------------
  // Parse the input database.  Unlike Do178Builder itself, we use libxml
  // to parse the input file.  That way the wheel isn't reinvented again
  // (since I already reinvented it once, this isn't redundant), and we get
  // an independent implementation to check our work.
  
  // Unfortunately, it turns out that the Do178Builder database XML is not
  // correct XML in that tag names aren't always formed correctly according to
  // the rules, so we need to translate it first.
  if (!Options.NoSed)
    {
      //sprintf (Command, "sed \"s/\\(&#\\)\\([0-9][0-9]\\)\\(;\\)/_r\\2r_/g\" <%s >%s",
      //       Options.DatabaseName, TempFilename);
      sprintf (Command, "sed " SED1 " <%s | sed " SED2 " | sed " SED3 " >%s",
      	       Options.DatabaseName, TempFilename);
      if (system (Command))
	{
	  fprintf (stderr, "Error: Attempt to run sed failed\n");
	  goto Done;
	}
      // Next, tag names are not allowed to begin with a digit, so we have to
      // seek out and destroy those.
      //	sed "" 
    }
  
  // Initialize libxml, and check the library version.
  LIBXML_TEST_VERSION
  
  // Parse the file into the memory structure 'doc'.
  doc = xmlReadFile(TempFilename, NULL, 0);
  if (doc == NULL) 
    {
      printf("Error: Failed to parse input XML\n");
      goto Done;
    }
    
  //----------------------------------------------------------------------------
  // Compile all of the regular-expression state-machines.
  if (Pattern (&Options.HighLabelPatternMachine, Options.HighLabelPattern, &Options))
    goto Done;
  if (Pattern (&Options.HighSectionPatternMachine, Options.HighSectionPattern, &Options))
    goto Done;
  if (Pattern (&Options.HighSectionRangeMachine, Options.HighSectionRange, &Options))
    goto Done;
  if (Pattern (&Options.LowLabelPatternMachine, Options.LowLabelPattern, &Options))
    goto Done;
  if (Pattern (&Options.LowSectionPatternMachine, Options.LowSectionPattern, &Options))
    goto Done;
  if (Pattern (&Options.LowSectionRangeMachine, Options.LowSectionRange, &Options))
    goto Done;
  if (Pattern (&Options.SystemListPatternMachine, SYSTEM_LIST_PATTERN, &Options))
    goto Done;
  if (Pattern (&Options.SystemSectionRangeMachine, Options.SystemSectionRange, &Options))
    goto Done;
  if (Pattern (&Options.VerifySectionRangeMachine, Options.VerifySectionRange, &Options))
    goto Done;
    
  //----------------------------------------------------------------------------
  // Do whatever processing was specified.
  
  if (Options.Display)
    {
      if (WalkNodes (xmlDocGetRootElement(doc), &PrintNodeName, &Options))
        {
	  fprintf (stderr, "XML display failed.\n");
	  goto Done;
	}
    }
    
  if (Options.TraceDesign || Options.TraceDesignFull || Options.TraceSystem ||
      Options.TraceImplement || Options.TraceVerify)
    {
      TraceInitialize (&Options);
      if (WalkNodes (xmlDocGetRootElement(doc), &Trace, &Options))
        {
	  fprintf (stderr, "Trace matrix failed.\n");
	  goto Done;
	}
      TraceCloseout (&Options);
    }
 
  //----------------------------------------------------------------------------
  // All done!
  RetVal = 0;
Done:
  return (RetVal); 
}


