/*
  Name:        	TraceDesignSystem.c
  Author:      	Ron Burkey
  Purpose:     	The processing for Do178Builder-utility 
  		that creates the traceability matrix for the
		--trace-system command-line switch.
  Mod History:	03/14/2008 RSB	Began.
  		03/18/2008 RSB	Adapted TraceDesign.c to serve
				this purpose.  What TraceDesign.c
				does is actually quite close (or
				quite analogous) to what's needed,
				but different enough that it can't
				be adapted as trivially as 
				TraceDesignFull.c was.
		03/20/2008 RSB	Added some subset matching to the
				system requirements.
		03/21/2008 RSB  Added requirements-document section
				title to output file.
		03/24/2008 RSB	Added "{*}".
  
  Copyright 2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  
  The dataset output by this function is a CSV file with one record 
  for each pairing of a system requirement with a high-level software
  (or hardware) requirement.  The SRD (or HRD) section number is also
  listed.  A system requirement which has no association is matched
  with "(none)", while a software requirement with no associated system
  requirement is matched with "(derived)".  

  The document sections eligible for including the system requirements
  traceability references are identified by pattern matching with a
  user-supplied regular expression. The traceability references to
  the system requirements are recognized within the eligible document
  sections because they have DocBook markup that looks like this:
  	<docbook>ReferenceList</docbook>
  The ReferenceList consists of one or more individual references
  separated by a delimiter (specified by --delimiter-system).  The 
  delimiter can be surrounded by optional white space.
  
  By sucking the CSV file into a spreadsheet program, it can be sorted 
  in order of the high-level requirements or in order of the low-level
  requirements.  The actual output order is simply the order in which
  the requirements are found in the XML.
*/

#include "Do178Builder-utility.h"

#define CsvFilename "Do178Builder-trace-system.csv"
// The reason the following lines are here is that this file was originally
// adapted from TraceDesign.c, and we want to make sure that we've changed
// the names of all data structures referring to --trace-design, so we
// reassign them to nonsense words that would result in compile-time
// errors.
#define TraceDesignState eeny
#define TraceDesignCloseout meeny
#define FileTraceDesign miny
#define TraceDesign moe

// A dummy string used for processing section titles.
static char Title[1024];

void 
TraceSystemCloseout (ProcessingOptions_t *Options)
{
  int k, m;
  char *ReqDocName;
  
  if (Options->CombineSrdSdd)
    ReqDocName = Options->ComDocName;
  else
    ReqDocName = Options->ReqDocName;
    
  // First do all of the system requirements marked with "{*}".
  for (k = 0; k < Options->TraceSystemState.NumSystemRequirements; k++)
    {
      int OkayToPrint = 0;
      
      // Check if there are any "{*}" marks:
      for (m = 0; m < Options->TraceSystemState.NumOneToOne[k]; m++)
	if (!strcmp ("*", Options->TraceSystemState.OneToOne[k][m]))
	  {
	    OkayToPrint = 1;
	    break;
	  }
  
      if (OkayToPrint)
        {
	  PrettifyTitle (Options->TraceSystemState.SystemRequirementSection.Title,
			 Title);
	  fprintf (Options->FileTraceSystem, "\"%s\",\"",
		   Options->TraceSystemState.SystemRequirements[k]);
	  for (m = 0; m < Options->TraceSystemState.NumHighLevelRequirements; m++)
	    {
	      if (m > 0)
	        fprintf (Options->FileTraceSystem, "\n");
	      fprintf (Options->FileTraceSystem, "%s",
		       Options->TraceSystemState.HighLevelRequirements[m]);
	    }
	  fprintf (Options->FileTraceSystem, "\",\"%s %s or below\",\"%s\"\n",
		   ReqDocName,
		   Options->TraceSystemState.SystemRequirementSection.Number,
		   Title);
	       
	  Options->TraceSystemState.HighLevelCount++;
        }
    }
  
  // All of the system requirements for which there were no high-level 
  // (hardware or software) requirements.
  if (!Options->TraceSystemState.HighLevelCount)
    {
      for (k = 0; k < Options->TraceSystemState.NumSystemRequirements; k++)
        {
	  PrettifyTitle (Options->TraceSystemState.SystemRequirementSection.Title,
	                 Title);
	  fprintf (Options->FileTraceSystem, "\"%s\",\"%s\",\"%s %s or below\",\"%s\"\n",
		   Options->TraceSystemState.SystemRequirements[k],
		   "(none)",
		   ReqDocName,
		   Options->TraceSystemState.SystemRequirementSection.Number,
		   Title);
        }
    }
  Options->TraceSystemState.NumSystemRequirements = 0;
  Options->TraceSystemState.HighLevelCount = 0;
  Options->TraceSystemState.NumHighLevelRequirements = 0;
}

// This function does the processing by being called by WalkNodes
// once for each node in the XML tree.  It has to maintain its own
// state machine in order to tie all of these separate calls together,
// and the state machine comprises fields in Options.  Calling with
// a_node==NULL should initialize (reset) the state machine.
int 
TraceSystem (xmlNode *a_node, ProcessingOptions_t *Options)
{
  int i, j, k, m,n;
  char c;
  int CanContainHighLevel = 0, CanContainSystem = 0;
  int DoesContainHighLevel = 0, DoesContainSystem = 0;
  regmatch_t pmatch[2];
  char *Text, *ReqDocName;
      
  // Initialize?
  if (a_node == NULL)
    {
      Options->FileTraceSystem = fopen (CsvFilename, "w");
      if (Options->FileTraceSystem == NULL)
        {
	  fprintf (stderr, "Error: Cannot create %s\n", CsvFilename);
	  return (1);
	}
      Options->TraceSystemState.NumSystemRequirements = 0;
      fprintf (Options->FileTraceSystem, 
      	       "\"System Requirement\","
               "\"High-Level Requirement\","
	       "\"Requirements Document Section\","
	       "\"Requirements Document Section Title\"\n");
      return (0);
    }
    
  if (Options->CombineSrdSdd)
    ReqDocName = Options->ComDocName;
  else
    ReqDocName = Options->ReqDocName;
    
  // Process the node. ---------------------------------------------
  
  // Determine if it can contain a system requirement.  (Not 
  // necessarily that it DOES, just that it COULD.)  This will
  // be so if:
  //	1. It is part of the requirements document; and
  //	2. It matches the section-number constraint from the command
  //	   line switches.
  // By default, it is marked as NOT possible, so we only need to 
  // check the positive cases.
  if (Options->DocSections[REQ_DOC].Present)
    {
      if (Options->SystemSectionRange == NULL)
        CanContainSystem = 1;	// No range constraint, so all okay!
      else
        {
	  // A section-number pattern has been defined by a command-line
	  // switch as a constraint on the sections that can contain
	  // requirements, so apply it and see if we're in the right
	  // range of section numbers.  Note that we make no effort
	  // to insure that the match encompasses the entire section
	  // number.  If the user desires that, then he should have
	  // prefixed and suffixed his pattern with '^' and '$' to 
	  // match the beginning of line and end of line.  If regexec
	  // returns 0, then it means a match was found.  
	  if (0 == regexec (&Options->SystemSectionRangeMachine,
	  		    Options->DocSections[REQ_DOC].Number,
			    1, pmatch, 0))
	    CanContainSystem = 1;
	}
    }

  // If we've found a section that could contain system requirements,
  // then we need to close out any existing system requirements that 
  // we may have been accumulating high-level requirements for.
  if (CanContainSystem)
    TraceSystemCloseout (Options);

  // If could contain system requirements, let's check to see if it
  // DOES contain any.  If so, we have to buffer them.  We can't do anything
  // with them immediately, since we can only write them out when we have
  // some high-level requirements to go with them.
  if (CanContainSystem)
    {
      Options->TraceSystemState.SystemRequirementSection = Options->DocSections[REQ_DOC];
      // Get the text of the node.
      Text = (char *) xmlNodeListGetString (a_node->doc, a_node->xmlChildrenNode, 1);
      printf ("SY candidate %s, section %s:", Options->ReqDocName, 
             Options->DocSections[REQ_DOC].Number);
      
      // Now determine if there are any matches for the requirements
      // language patterns.  This is a little different from what goes
      // on in places like TraceDesign.c, since there we only are interested
      // in the entire match (pmatch[0]), whereas here we're interested in
      // the match for the first paranthesized sub-expression (pmatch[1]).
      for (i = j = 0; !regexec (&Options->SystemListPatternMachine, &Text[j], 2, pmatch, 0); j += pmatch[0].rm_eo)
        {
	  pmatch[1].rm_so += j;		// Turn relative offsets into absolute.
	  pmatch[1].rm_eo += j;
	  
	  // Now unlike TraceDesign.c and friends, the text we've identified 
	  // not the system requirement.  Instead, it's (potentially at least)
	  // a list of system requirements separated by delimiters and optional
	  // whitespace.  So we have to pick off the individual system reqs.
	  // Moreover, each individual system req. may be preceded by up to
	  // MAX_ONE_TO_ONE expressions of the form
	  //	{S}
	  // where S is a string that's supposed to match one of the software
	  // requirements or hardware requirements. 
	  c = Text[pmatch[1].rm_eo];
	  Text[pmatch[1].rm_eo] = 0;
	  //printf (" List=\"%s\"", &Text[pmatch[1].rm_so]);
	  for (k = pmatch[1].rm_so; k < pmatch[1].rm_eo; )
	    {
	      char *s, *ss;
	      char c, cc;
	      // Move past any leading spaces.
	      for (; k < pmatch[1].rm_eo && isspace (Text[k]); k++);
	      if (k >= pmatch[1].rm_eo)
	        break;		// At end of string.
	      s = strchr (&Text[k], Options->DelimiterSystem);
	      if (s == NULL)
	        s = strlen (&Text[k]) + &Text[k];
		
              // At this point, &Text[k] points to a nul-terminated requirement,
	      // possibly with trailing whitspace at the end which needs to be 
	      // trimmed.  We are assured that it's not ALL whitespace.
	      for (ss = s; isspace (ss[-1]); ss--);
	      // Nul-terminate it for saving.
	      cc = *ss;
	      *ss = 0;
	      // Will there be enough room to save this?
	      if (i >= MAX_MATCHES_PER_NODE)
		{
		  fprintf (stderr, "Error: Too many requirements in a section.\n");
		  xmlFree (Text);
		  return (1);
		}
	      // Now, pick off the {S} thingies, if any.
	      Options->TraceSystemState.NumOneToOne[i] = 0;
	      while (Text[k] == '{')
	        {
		  char *ss;
		  int n;
		  if (Options->TraceSystemState.NumOneToOne[i] >= MAX_ONE_TO_ONE)
		    {
		      fprintf (stderr, "Error: Too many {} expressions.\n");
		      xmlFree (Text);
		      return (1);
		    }  
		  for (ss = &Text[k + 1]; *ss && *ss != '}'; ss++);
		  if (*ss != '}')
		    {
		      fprintf (stderr, "Error: Unterminated {} expression.\n");
		      xmlFree (Text);
		      return (1);
		    } 
		  n = ss - &Text[k + 1];
		  if (n >= SHORT_REQUIREMENT_SIZE)
		    {
		      fprintf (stderr, "Error: {} expression too long.\n");
		      xmlFree (Text);
		      return (1);
		    }
		  *ss = 0;
		  strcpy (Options->TraceSystemState.OneToOne[i][Options->TraceSystemState.NumOneToOne[i]], &Text[k + 1]);
		  Options->TraceSystemState.NumOneToOne[i]++;
		  *ss = '}';
		  k += n + 2;
		}
	      // Helpful message.
	      for (m = 0; m < Options->TraceSystemState.NumOneToOne[i]; m++)
	        printf (" {%s}", Options->TraceSystemState.OneToOne[i][m]);
	      printf (" \"%s\"", &Text[k]);
	      // Save the requirement.
	      n = ss - &Text[k];
	      if (n >= MAX_REQUIREMENT_SIZE)
		{
		  fprintf (stderr, "Error: Pattern produces string length >%d\n", MAX_REQUIREMENT_SIZE - 1);
		  xmlFree (Text);
		  return (1);
		}
	      strcpy (Options->TraceSystemState.SystemRequirements[i], &Text[k]);
	      i++;
	      // Undo the nul-termination.
	      *ss = cc;		
		
	      // This requirement has been processed, so move on the next one
	      // in the list.	
	      if (*s == 0)
	        break;		// At end of string.
	      *s = c;
	      k = s - &Text[0] + 1;
	    }
	  Text[pmatch[1].rm_eo] = c;
	}
	
      Options->TraceSystemState.NumSystemRequirements = i;
      if (i == 0)
        {
          printf (" (none)");
	}
      //else
      //  for (i = 0; i < Options->TraceSystemState.NumSystemRequirements; i++)
      //    printf (" %s", Options->TraceSystemState.SystemRequirements[i]);
      printf ("\n");
      xmlFree (Text);      
    }

  // Similarly, for the high-level requirements.  
  if (Options->DocSections[REQ_DOC].Present)
    {
      if (Options->HighSectionRange == NULL)
        CanContainHighLevel = 1;	// No range constraint, so all okay!
      else
        {
	  if (0 == regexec (&Options->HighSectionRangeMachine,
	  		    Options->DocSections[REQ_DOC].Number,
			    1, pmatch, 0))
	    CanContainHighLevel = 1;
	}
    }
    
  if (CanContainHighLevel)
    {
      char *s;
      // Get the text of the node.
      Text = (char *) xmlNodeListGetString (a_node->doc, a_node->xmlChildrenNode, 1);
      printf("HL candidate %s, section %s:", Options->ReqDocName, 
             Options->DocSections[REQ_DOC].Number, Text);
      
      // Now determine if there are any matches for the requirements
      // language patterns.  We only check the so-called "section" language
      // matches if there are no "label" matches.
      for (i = j = 0; !regexec (&Options->HighLabelPatternMachine, &Text[j], 1, pmatch, 0); i++, j = pmatch[0].rm_eo)
        {
	  pmatch[0].rm_so += j;		// Turn relative offsets into absolute.
	  pmatch[0].rm_eo += j;
	  n = pmatch[0].rm_eo - pmatch[0].rm_so;
	  c = Text[pmatch[0].rm_eo];
	  Text[pmatch[0].rm_eo] = 0;
	  printf (" %s", &Text[pmatch[0].rm_so]);
	  
	  // Will there be enough room to save this?
	  if (i >= MAX_MATCHES_PER_NODE)
	    {
	      fprintf (stderr, "Error: Too many requirements in a section.\n");
	      xmlFree (Text);
	      return (1);
	    }
	  if (n >= MAX_REQUIREMENT_SIZE)
	    {
	      fprintf (stderr, "Error: Requirement string length >%d\n", MAX_REQUIREMENT_SIZE - 1);
	      xmlFree (Text);
	      return (1);
	    }
	  strcpy (Options->TraceSystemState.HighLevelRequirements[i], &Text[pmatch[0].rm_so]);
	  Options->TraceSystemState.NumHighLevelRequirements++;
	  
	  for (k = 0; k < Options->TraceSystemState.NumSystemRequirements; k++)
	    {
	      // Before outputting the record, we need to check the system 
	      // requirement against the list of valid software requirements, if any.
	      int OkayToPrint = 0;
	      if (Options->TraceSystemState.NumOneToOne[k] == 0)
	        OkayToPrint = 1;
	      else
	        {
		  for (m = 0; m < Options->TraceSystemState.NumOneToOne[k]; m++)
		    if (!strcmp (&Text[pmatch[0].rm_so],
		    		 Options->TraceSystemState.OneToOne[k][m]))
		      {
		        OkayToPrint = 1;
			break;
		      }
		}
	      if (OkayToPrint)	
	        {
		  PrettifyTitle (Options->TraceSystemState.SystemRequirementSection.Title,
				 Title);
		  fprintf (Options->FileTraceSystem, "\"%s\",\"%s %s\",\"%s %s\",\"%s\"\n",
			   Options->TraceSystemState.SystemRequirements[k],
			   ReqDocName,
			   &Text[pmatch[0].rm_so],
			   ReqDocName,
			   Options->DocSections[REQ_DOC].Number,
			   Title);
	        }
            }		     
	  if (k == 0)
	    {
	      PrettifyTitle (Options->TraceSystemState.SystemRequirementSection.Title,
			     Title);
	      fprintf (Options->FileTraceSystem, "\"%s\",\"%s %s\",\"%s %s\",\"%s\"\n",
		       "(derived)",
		       ReqDocName,
		       &Text[pmatch[0].rm_so],
		       ReqDocName,
		       Options->DocSections[REQ_DOC].Number,
		       Title);
	    }
	  Text[pmatch[0].rm_eo] = c;
	}
      for (; !i && !regexec (&Options->HighSectionPatternMachine, Text, 1, pmatch, 0); i++)
        {
	  printf (" %s", Options->DocSections[REQ_DOC].Number);
	  for (k = 0; k < Options->TraceSystemState.NumSystemRequirements; k++)
	    {
	      PrettifyTitle (Options->TraceSystemState.SystemRequirementSection.Title,
			     Title);
	      fprintf (Options->FileTraceSystem, "\"%s\",\"%s %s\",\"%s %s\",\"%s\"\n",
		       Options->TraceSystemState.SystemRequirements[k],
		       ReqDocName,
		       Options->DocSections[REQ_DOC].Number,
		       ReqDocName,
		       Options->DocSections[REQ_DOC].Number,
		       Title);
	    }
	  if (k == 0)
	    {
	      PrettifyTitle (Options->TraceSystemState.SystemRequirementSection.Title,
			     Title);
	      fprintf (Options->FileTraceSystem, "\"%s\",\"%s %s\",\"%s %s\",\"%s\"\n",
		       "(derived)",
		       ReqDocName,
		       Options->DocSections[REQ_DOC].Number,
		       ReqDocName,
		       Options->DocSections[REQ_DOC].Number,
		       Title);
	    }
	}
      if (i == 0)
        {
          printf (" (none)");
	}
      printf ("\n");
      
      Options->TraceSystemState.HighLevelCount += i;
      
      xmlFree(Text);      
    }

  return (0);
}



