///////////////////////////////////////////////////////////////////////////
// Filename:    XmlEditorDialogClass.cpp
// Author:      Ron Burkey
// Purpose:     The code for the XmlEditorDialogClass C++ class.
// Mods:        10/03/01 RSB    Split off from Do178Builder.cpp.
//              10/21/01 RSB    Added code for the find and find-again
//                              buttons.
//              10/28/01 RSB    Began adding stuff for italics, bold, underlining,S
//                              and fixed-width. ... but it turns out that the
//                              rich-text stuff I thought had been implemented
//                              in wxWindows hasn't been, so none of it does
//                              anything.
//              10/31/01 RSB    Removed Italics, bold, underline, and code
//                              markup (which were incomplete and not working
//                              anyhow), in favor of the beginnings of a
//                              general DocBook markup feature.
//              02/18/02 RSB    Tried fixing the markup-insertion bug in Win32.
//                              Made some attempt to fix the Win32 button
//                              non-refresh problem.  Tried to fix the resizing
//                              problem in Win32.
//              08/06/03 RSB    Added shortcut toolbuttons for Emphasis, Function,
//                              Application, and ProgramListing markup.
//              03/26/04 RSB    Added shortcut toolbuttons for raw-DocBook and 
//                              for document-builds.
//              03/27/04 RSB    For some reason, CreateTreeCopy was balled up so
//                              as to include only one child at any given level.
//                              The correct code was in place, but was commented
//                              out.  There's undoubtedly some reason for this,
//                              but I can't fathom what it must have been,
//                              because the comments don't say.
//              04/17/04 RSB    Added the <trace> hotbutton.
//              06/23/06 RSB    Began adapting for DO-254.
//              06/30/06 RSB    Added doc-back and doc-next.
//              07/01/06 RSB    Added InsertGraphic, InsertList, InsertTable.
//              07/02/06 RSB    Fixed ampersand in program listing.
//		07/03/06 RSB	Tried to implement FindAgain.
//		01/26/08 RSB	Adjusted the way various DocBook commands are
//				inserted, so that they are consistent with
//				both DocBook SGML and XML rather than taking
//				advantage of the shortcuts allowed in SGML.
//		03/23/08 RSB	The checkboxes for determining which document
//				types include a given document section are now
//				always changeable by the user without having 
//				to enter framework-editing mode.  Added support 
//				for new toolbar buttons (AddCustom, AddSubCustom,
//				InsertHyperlink, InsertFootnote, InsertMdash).
//				Fixed the focus on the various toolbar buttons
// 				for inserting stuff like graphics, tables, 
//				etc., so that the focus returns to the 
//				text.  Also, for some of these the dataset-changed
//				flag wasn't being set.  Fixed the annoying 
//				problem with 'search', in that the found item
//				often wasn't highlighted, because focus wasn't
//				returned to that window.
//		03/24/08 RSB	wxWidgets (both 2.4 and 2.6 as far as I can tell)
//				seem to have a bug in which SetSelection for a
//				TextCtrl may or may not work.  SetSelection(n,n)
//				always seems to work, but SetSelection(n,m) for
//				n!=m may not.
//		03/26/08 RSB	Fixed up for wxWidgets 2.8.
//		04/05/08 RSB	Added nice indenting when inserting graphic,
//				table, or list templates.  Added a toolbutton
//				for building the documents, so that you don't
//				have to leave the editor to do that.
//		04/07/08 RSB	Fixed an initialization error for LastChild
//				in the CreateTreeCopy function.  (I don't know
//				if the error is real or merely apparent, but 
//				the fix should be fine in either case.)

/*
  Copyright 2001-2004,2006,2008 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
  The XmlEditorDialogClass implements the dialog in the Do178Builder
  program used for editing the data for a Do178Builder project.
  This is basically all of the code for the class, except as follows:
  
	Do178Builder.h		Contains the prototype for the class
				and for the XmlStruct structure, along
				with unrelated items.
	Do178Builder.cpp	Contains the wxWindows event table for
				the dialog, and items unrelated to 
				the class.
	Do178Builder_wdr.cpp	Contains functions for constructing 
				the dialog from wxWindows widgets,
				functions providing pointers to all
				relevant bitmaps, and various items
				unrelated to the class.
				
  The reason for this (perhaps) rather odd partitioning of the code is
  that the wxDesigner RAD environment was used to construct the user 
  interface, and it arranges the code in this way.  It also places all
  of the event handlers for the class into Do178Builder.cpp, but these
  have been moved to XmlEditorDialogClass.cpp manually.  
*/

/////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation
#include "wx/wxprec.h"

#if wxMAJOR_VERSION > 2 || wxMINOR_VERSION > 4
#include <wx/numdlg.h>
#endif

// Include private headers
#include "Do178Builder.h"

#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
extern bool HintEditable;
extern bool TemporaryHintEditable;
extern int DatasetChanged;
extern wxString rFilename;
extern wxTreeItemId INVALID_TREE_ITEM;


//---------------------------------------------------------------------------
// The constructor for the class.

XmlEditorDialogClass::XmlEditorDialogClass (wxWindow *
					    parent,
					    wxWindowID id,
					    const wxString &
					    title,
					    const wxPoint &
					    position,
					    const wxSize & size, long style):
wxDialog (parent, id, title, position, size, style)
{
  extern int Do254, FontSize, TreeFontSize;
  int i, DoctypeEditable;
  XmlEditorDialog (this, TRUE);

  // The color used to mark data items changed.
  MY_CHANGE_COLOR = wxColour (0, 0, 255);
  CHANGE_COLOR = &MY_CHANGE_COLOR;

  // Regenerate pointers to all of the dialog elements we need to address directly.
  DocChoice = GetDocchoice ();
  MarkupChoice = GetMarkupchoice ();
  MarkupButton = GetMarkupbutton ();
  NoviceButton = GetNovicebutton ();
  ExpertButton = GetExpertbutton ();
  ItemTree = GetXmleditor ();	//wxDynamicCast (FindWindow ("ItemTree"), wxTreeCtrl);
  HintBox = GetHinttext ();	//wxDynamicCast (FindWindow ("HintBox"), wxTextCtrl);
  ItemText = GetItemtext ();	//wxDynamicCast (FindWindow ("ItemText"), wxTextCtrl);
  PsacCheck = GetPsaccheck ();	//wxDynamicCast (FindWindow ("PsacCheck"), wxCheckBox);
  SdpCheck = GetSdpcheck ();	//wxDynamicCast (FindWindow ("SdpCheck"), wxCheckBox);
  SvpCheck = GetSvpcheck ();	//wxDynamicCast (FindWindow ("SvpCheck"), wxCheckBox);
  ScmpCheck = GetScmpcheck ();	//wxDynamicCast (FindWindow ("ScmpCheck"), wxCheckBox);
  SqapCheck = GetSqapcheck ();	//wxDynamicCast (FindWindow ("SqapCheck"), wxCheckBox);
  SrsCheck = GetSrscheck ();	//wxDynamicCast (FindWindow ("SrsCheck"), wxCheckBox);
  SdsCheck = GetSdscheck ();	//wxDynamicCast (FindWindow ("SdsCheck"), wxCheckBox);
  ScsCheck = GetScscheck ();	//wxDynamicCast (FindWindow ("ScsCheck"), wxCheckBox);
  SrdCheck = GetSrdcheck ();	//wxDynamicCast (FindWindow ("SrdCheck"), wxCheckBox);
  DdCheck = GetDdcheck ();	//wxDynamicCast (FindWindow ("DdCheck"), wxCheckBox);
  SvcpCheck = GetSvcpcheck ();	//wxDynamicCast (FindWindow ("SvcpCheck"), wxCheckBox);
  SeciCheck = GetSecicheck ();	//wxDynamicCast (FindWindow ("SeciCheck"), wxCheckBox);
  SciCheck = GetScicheck ();	//wxDynamicCast (FindWindow ("SciCheck"), wxCheckBox);
  SasCheck = GetSascheck ();	//wxDynamicCast (FindWindow ("SasCheck"), wxCheckBox);
  PsacSection = GetPsacsection ();	//wxDynamicCast (FindWindow ("PsacSection"), wxCheckBox);
  SdpSection = GetSdpsection ();	//wxDynamicCast (FindWindow ("SdpSection"), wxCheckBox);
  SvpSection = GetSvpsection ();	//wxDynamicCast (FindWindow ("SvpSection"), wxCheckBox);
  ScmpSection = GetScmpsection ();	//wxDynamicCast (FindWindow ("ScmpSection"), wxCheckBox);
  SqapSection = GetSqapsection ();	//wxDynamicCast (FindWindow ("SqapSection"), wxCheckBox);
  SrsSection = GetSrssection ();	//wxDynamicCast (FindWindow ("SrsSection"), wxCheckBox);
  SdsSection = GetSdssection ();	//wxDynamicCast (FindWindow ("SdsSection"), wxCheckBox);
  ScsSection = GetScssection ();	//wxDynamicCast (FindWindow ("ScsSection"), wxCheckBox);
  SrdSection = GetSrdsection ();	//wxDynamicCast (FindWindow ("SrdSection"), wxCheckBox);
  DdSection = GetDdsection ();	//wxDynamicCast (FindWindow ("DdSection"), wxCheckBox);
  SvcpSection = GetSvcpsection ();	//wxDynamicCast (FindWindow ("SvcpSection"), wxCheckBox);
  SeciSection = GetSecisection ();	//wxDynamicCast (FindWindow ("SeciSection"), wxCheckBox);
  SciSection = GetScisection ();	//wxDynamicCast (FindWindow ("SciSection"), wxCheckBox);
  SasSection = GetSassection ();	//wxDynamicCast (FindWindow ("SasSection"), wxCheckBox);
  BitmapSatA = GetSata ();	//wxDynamicCast (FindWindow ("BitmapSatA"), wxBitmapButton);
  BitmapSatB = GetSatb ();	//wxDynamicCast (FindWindow ("BitmapSatB"), wxBitmapButton);
  BitmapSatC = GetSatc ();	//wxDynamicCast (FindWindow ("BitmapSatC"), wxBitmapButton);
  BitmapSatD = GetSatd ();	//wxDynamicCast (FindWindow ("BitmapSatD"), wxBitmapButton);
  BitmapSatE = GetSate ();	//wxDynamicCast (FindWindow ("BitmapSatE"), wxBitmapButton);
  CcA = GetCca ();		//wxDynamicCast (FindWindow ("CcA"), wxBitmapButton);
  CcB = GetCcb ();		//wxDynamicCast (FindWindow ("CcB"), wxBitmapButton);
  CcC = GetCcc ();		//wxDynamicCast (FindWindow ("CcC"), wxBitmapButton);
  CcD = GetCcd ();		//wxDynamicCast (FindWindow ("CcD"), wxBitmapButton);
  CcE = GetCce ();		//wxDynamicCast (FindWindow ("CcE"), wxBitmapButton);
  Add = GetAddItem ();		//wxDynamicCast (FindWindow ("Add"), wxButton);
  AddSub = GetSubitem ();	//wxDynamicCast (FindWindow ("AddSub"), wxButton);
  AddCustom = GetAddItemCustom ();
  AddSubCustom = GetSubitemCustom ();
  Del = GetDeleteItem ();	//wxDynamicCast (FindWindow ("Del"), wxButton);
  AppLevel = GetAppLevel ();
  AppA = GetAppA ();
  AppB = GetAppB ();
  AppC = GetAppC ();
  AppD = GetAppD ();
  AppE = GetAppE ();
  SeparatorLine = GetLine ();
  InsertTable = GetInsertTable ();
  InsertGraphic = GetInsertGraphic ();
  InsertList = GetInsertList ();
  InsertHyperlink = GetInsertHyperlink ();
  InsertFootnote = GetInsertFootnote ();
  InsertMdash = GetInsertMdash ();

  InsertTable->Enable (true);
  InsertList->Enable (true);
  InsertGraphic->Enable (true);

  // Adapt for DO-254 vs. DO-178B
  if (Do254)
    {
      PsacCheck->SetLabel ("PHAC");
      PsacSection->SetLabel ("PHAC");
      SdpCheck->SetLabel ("HDP");
      SdpSection->SetLabel ("HDP");
      SvcpCheck->SetLabel ("HTPR");
      SvcpSection->SetLabel ("HTPR");
      SrdCheck->SetLabel ("HRD");
      SrdSection->SetLabel ("HRD");
      DdCheck->SetLabel ("HDRD");
      DdSection->SetLabel ("HDRD");
      SvpCheck->SetLabel ("HVVP");
      SvpSection->SetLabel ("HVVP");
      ScmpCheck->SetLabel ("HCMP");
      ScmpSection->SetLabel ("HCMP");
      SqapCheck->SetLabel ("HPAP");
      SqapSection->SetLabel ("HPAP");
      SrsCheck->SetLabel ("RS");
      SrsSection->SetLabel ("RS");
      SdsCheck->SetLabel ("HDS");
      SdsSection->SetLabel ("HDS");
      ScsCheck->SetLabel ("VVS");
      ScsSection->SetLabel ("VVS");
      SeciCheck->SetLabel ("HATC");
      SeciSection->SetLabel ("HATC");
      SciCheck->SetLabel ("HARS");
      SciSection->SetLabel ("HARS");
      SasCheck->SetLabel ("HAS");
      SasSection->SetLabel ("HAS");
      AppLevel->Hide ();
      AppA->Hide ();
      AppB->Hide ();
      AppC->Hide ();
      AppD->Hide ();
      AppE->Hide ();
      BitmapSatA->Hide ();
      BitmapSatB->Hide ();
      BitmapSatC->Hide ();
      BitmapSatD->Hide ();
      BitmapSatE->Hide ();
      SeparatorLine->Hide ();
      DocChoice->Append ("PHAC");
      DocChoice->Append ("HDP");
      DocChoice->Append ("HVVP");
      DocChoice->Append ("HCMP");
      DocChoice->Append ("HPAP");
      DocChoice->Append ("RS");
      DocChoice->Append ("HDS");
      DocChoice->Append ("VVS");
      DocChoice->Append ("HRD");
      DocChoice->Append ("HDRD");
      DocChoice->Append ("HTPR");
      DocChoice->Append ("HATC");
      DocChoice->Append ("HARS");
      DocChoice->Append ("HAS");
    }
  else
    {
      PsacCheck->SetLabel ("PSAC");
      PsacSection->SetLabel ("PSAC");
      SdpCheck->SetLabel ("SDP");
      SdpSection->SetLabel ("SDP");
      SvpCheck->SetLabel ("SVP");
      SvpSection->SetLabel ("SVP");
      SrdCheck->SetLabel ("SRD");
      SrdSection->SetLabel ("SRD");
      DdCheck->SetLabel ("SDD");
      DdSection->SetLabel ("SDD");
      SvcpCheck->SetLabel ("SVCP");
      SvcpSection->SetLabel ("SVCP");
      ScmpCheck->SetLabel ("SCMP");
      ScmpSection->SetLabel ("SCMP");
      SqapCheck->SetLabel ("SQAP");
      SqapSection->SetLabel ("SQAP");
      SrsCheck->SetLabel ("SRS");
      SrsSection->SetLabel ("SRS");
      SdsCheck->SetLabel ("SDS");
      SdsSection->SetLabel ("SDS");
      ScsCheck->SetLabel ("SCS");
      ScsSection->SetLabel ("SCS");
      SeciCheck->SetLabel ("SECI");
      SeciSection->SetLabel ("SECI");
      SciCheck->SetLabel ("SCI");
      SciSection->SetLabel ("SCI");
      SasCheck->SetLabel ("SAS");
      SasSection->SetLabel ("SAS");
      DocChoice->Append ("PSAC");
      DocChoice->Append ("SDP");
      DocChoice->Append ("SVP");
      DocChoice->Append ("SCMP");
      DocChoice->Append ("SQAP");
      DocChoice->Append ("SRS");
      DocChoice->Append ("SDS");
      DocChoice->Append ("SCS");
      DocChoice->Append ("SRD");
      DocChoice->Append ("SDD");
      DocChoice->Append ("SVCP");
      DocChoice->Append ("SECI");
      DocChoice->Append ("SCI");
      DocChoice->Append ("SAS");
    }
  DocChoice->SetSelection (0);
  if (TreeFontSize < 0)
    ItemTree->SetFont (wxFont (-TreeFontSize, wxSWISS, wxNORMAL, wxNORMAL));
  else
    ItemTree->SetFont (wxFont (TreeFontSize, wxSWISS, wxNORMAL, wxBOLD));
  if (FontSize < 0)
    ItemText->SetFont (wxFont (-FontSize, wxMODERN, wxNORMAL, wxNORMAL));
  else
    ItemText->SetFont (wxFont (FontSize, wxMODERN, wxNORMAL, wxBOLD));

  // Enable the checkboxes, if appropriate.
  DoctypeEditable = 1; // Was =HintEditable until 20080323.
  PsacCheck->Enable (DoctypeEditable);
  SdpCheck->Enable (DoctypeEditable);
  SvpCheck->Enable (DoctypeEditable);
  ScmpCheck->Enable (DoctypeEditable);
  SqapCheck->Enable (DoctypeEditable);
  SrsCheck->Enable (DoctypeEditable);
  SdsCheck->Enable (DoctypeEditable);
  ScsCheck->Enable (DoctypeEditable);
  SrdCheck->Enable (DoctypeEditable);
  DdCheck->Enable (DoctypeEditable);
  SvcpCheck->Enable (DoctypeEditable);
  SeciCheck->Enable (DoctypeEditable);
  SciCheck->Enable (DoctypeEditable);
  SasCheck->Enable (DoctypeEditable);
  
  // ... and the other types of status boxes.
  BitmapSatA->Enable (HintEditable);
  BitmapSatB->Enable (HintEditable);
  BitmapSatC->Enable (HintEditable);
  BitmapSatD->Enable (HintEditable);
  BitmapSatE->Enable (HintEditable);
  CcA->Enable (HintEditable);
  CcB->Enable (HintEditable);
  CcC->Enable (HintEditable);
  CcD->Enable (HintEditable);
  CcE->Enable (HintEditable);

  // Default search parameters.
  SearchString = "";
  for (i = 0; i < NUM_DOCUMENTS; SearchDocs[i++] = 0);
  SearchAll = 1;
  SearchTitles = 0;
  SearchHints = 0;
  SearchText = 1;
  LastSearchPos = INVALID_TREE_ITEM;

  // Miscellaneous stuff.
  HintBox->SetEditable (HintEditable);
  wxColour HintBackgroundColor;
  if (HintEditable)
    HintBackgroundColor = *wxWHITE;
  else
    HintBackgroundColor = *wxLIGHT_GREY;
  HintBox->SetBackgroundColour (HintBackgroundColor);
  LastSelected = INVALID_TREE_ITEM;
  DatasetNewlyChanged = 0;
  ClickTreeBypass = 0;
  wxCommandEvent DummyEvent;
  NoviceCheck (DummyEvent);
}

//-------------------------------------------------------------
// Handlers for the Novice and Expert buttons.

extern const char *InlineTags[], *NoviceTags[];
extern int NumAllInlines, NumNovices;

void
XmlEditorDialogClass::ExpertCheck (wxCommandEvent & event)
{
  int i;
  wxChoice *MarkupChoice = GetMarkupchoice ();	//wxDynamicCast (FindWindow (ID_MARKUPCHOICE), wxChoice);
  MarkupChoice->Clear ();
  for (i = 0; i < NumAllInlines; i++)
    MarkupChoice->Append (InlineTags[i], (void *) NULL);
  MarkupChoice->SetColumns (3);
}

void
XmlEditorDialogClass::NoviceCheck (wxCommandEvent & event)
{
  int i;
  wxChoice *MarkupChoice = GetMarkupchoice ();	//wxDynamicCast (FindWindow (ID_MARKUPCHOICE), wxChoice);
  MarkupChoice->Clear ();
  for (i = 0; i < NumNovices; i++)
    MarkupChoice->Append (NoviceTags[i], (void *) NULL);
  MarkupChoice->SetColumns (1);
}

//-----------------------------------------------------------------
// Handler for the Markup button.

void
XmlEditorDialogClass::MarkupButtonClick (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  wxChoice *MarkupChoice = GetMarkupchoice ();	//wxDynamicCast (FindWindow (ID_MARKUPCHOICE), wxChoice);
  Tag = MarkupChoice->GetStringSelection ();
  if (Tag)
    {
      s = "</" + Tag + ">";
      //ItemText->Replace (to, to, s);
      ItemText->SetSelection (to, to);
      ItemText->SetInsertionPoint (to);
      ItemText->WriteText (s);
      s = "<" + Tag + ">";
      //ItemText->Replace (from, from, s);
      ItemText->SetSelection (from, from);
      ItemText->SetInsertionPoint (from);
      ItemText->WriteText (s);
      ItemText->SetSelection (from + 2 + Tag.Length (),
			      to + 2 + Tag.Length ());
      DatasetChanged = 1;
      ItemText->SetFocus ();
    }
}

// ... and hotbuttons for various markup choices.

void
XmlEditorDialogClass::ClickListingButton (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  // Take care of ampersands.
  for (n = from; n < to;)
    if (ItemText->GetValue ().GetChar (n) == '&')
      {
	ItemText->SetSelection (n + 1, n + 1);
	ItemText->SetInsertionPoint (n + 1);
	ItemText->WriteText ("amp;");
	n += 5;
	to += 4;
      }
    else
      n++;
  s = "\n</programlisting>\n";
  //ItemText->Replace (to, to, s);
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  s = "\n<programlisting>\n";
  //ItemText->Replace (from, from, s);
  ItemText->SetSelection (from, from);
  ItemText->SetInsertionPoint (from);
  ItemText->WriteText (s);
  ItemText->SetSelection (from + 18, to + 18);
  DatasetChanged = 1;
  ItemText->SetFocus ();
}

#if 0
// This shouldn't be necessary, but for some reason I'm finding 
// that SetSelection doesn't take!
void 
XmlEditorDialogClass::SetAndCheckSelection (long from, long to)
{
  //int i;
  long afrom, ato;
  printf ("From=%ld To=%ld\n", from, to);
  //ItemText->SetFocus ();
  //ItemText->SetSelection (1, 2);
  //ItemText->SetInsertionPoint (from);
  //wxMilliSleep (100);
  //for (i = 0; i < 10; i++)
  //  {
  //    ItemText->SetSelection (from, to);
  //    ItemText->GetSelection (&afrom, &ato);
  //    if (from == afrom && to == ato)
  //      break;
  //  }
  ItemText->SetSelection (from, to);
  ItemText->SetSelection (1, 5);
  ItemText->GetSelection (&afrom, &ato);
  printf ("Actual From=%ld, To=%ld\n", afrom, ato);
  //printf ("Actual i=%d From=%ld, To=%ld\n", i, afrom, ato);
  ItemText->SetInsertionPoint (from);
}
#endif // 0

void
XmlEditorDialogClass::ClickCodeButton (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  // Take care of ampersands.
  for (n = from; n < to;)
    if (ItemText->GetValue ().GetChar (n) == '&')
      {
	ItemText->SetSelection (n + 1, n + 1);
	ItemText->SetInsertionPoint (n + 1);
	ItemText->WriteText ("amp;");
	n += 5;
	to += 4;
      }
    else
      n++;
  s = "</function>";
  //ItemText->Replace (to, to, s);
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  s = "<function>";
  //ItemText->Replace (from, from, s);
  ItemText->SetSelection (from, from);
  ItemText->SetInsertionPoint (from);
  ItemText->WriteText (s);
  ItemText->SetSelection (from + 10, to + 10);
  //SetAndCheckSelection (from + 10, to + 10);
  DatasetChanged = 1;
  ItemText->SetFocus ();
}

void
XmlEditorDialogClass::ClickCommandButton (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  s = "</command>";
  //ItemText->Replace (to, to, s);
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  s = "<command>";
  //ItemText->Replace (from, from, s);
  ItemText->SetSelection (from, from);
  ItemText->SetInsertionPoint (from);
  ItemText->WriteText (s);
  ItemText->SetSelection (from + 9, to + 9);
  DatasetChanged = 1;
  ItemText->SetFocus ();
}

void
XmlEditorDialogClass::ClickEmphasisButton (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  s = "</emphasis>";
  //ItemText->Replace (to, to, s);
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  s = "<emphasis>";
  //ItemText->Replace (from, from, s);
  ItemText->SetSelection (from, from);
  ItemText->SetInsertionPoint (from);
  ItemText->WriteText (s);
  ItemText->SetSelection (from + 10, to + 10);
  DatasetChanged = 1;
  ItemText->SetFocus ();
}

// Here's a little thing to check if we're already between a
// <docbook> and </docbook tag pair.  This should be called with
// all text from the beginning through the insertion point.
int
InsideDocbook (wxString Text)
{
  int n;
  Text = Text.MakeLower ();
  while (1)
    {
      n = Text.Find ("<docbook>");
      if (n == -1)
	return (0);
      Text = Text.Mid (n + 9);
      n = Text.Find ("</docbook>");
      if (n == -1)
	return (1);
      Text = Text.Mid (n + 10);
    }
}

// Yes, I know it's not "DockBook".  The problem with code generators
// is that one little typo can propagate everywhere, and it's just easier
// to leave the misprint than to correct it.
void
XmlEditorDialogClass::DockBookButtonClick (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  if (InsideDocbook (ItemText->GetValue ().Truncate (to)))
    return;
  s = "</docbook>";
  //ItemText->Replace (to, to, s);
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  s = "<docbook>";
  //ItemText->Replace (from, from, s);
  ItemText->SetSelection (from, from);
  ItemText->SetInsertionPoint (from);
  ItemText->WriteText (s);
  ItemText->SetSelection (from + 9, to + 9);
  DatasetChanged = 1;
  ItemText->SetFocus ();
}

void
XmlEditorDialogClass::TraceClick (wxCommandEvent & event)
{
  long from, to, n;
  wxString Tag, s;
  ItemText->GetSelection (&from, &to);
  if (from == to)
    return;
  if (from > to)
    {
      n = from;
      from = to;
      to = n;
    }
  s = "</trace>";
  //ItemText->Replace (to, to, s);
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  s = "<trace>";
  //ItemText->Replace (from, from, s);
  ItemText->SetSelection (from, from);
  ItemText->SetInsertionPoint (from);
  ItemText->WriteText (s);
  ItemText->SetSelection (from + 7, to + 7);
  DatasetChanged = 1;
  ItemText->SetFocus ();
}

void XmlEditorDialogClass::AddSubitemCustomClick( wxCommandEvent &event )
{
  TemporaryHintEditable = TRUE;
  AddSubItem (event);
  TemporaryHintEditable = FALSE;
}

void XmlEditorDialogClass::AddItemCustomClick( wxCommandEvent &event )
{
  TemporaryHintEditable = TRUE;
  AddItem (event);
  TemporaryHintEditable = FALSE;
}

//-----------------------------------------------------------------
// A handler that gets called when the "find" button is pressed.

void
XmlEditorDialogClass::FindButton (wxCommandEvent & event)
{
  SearchBoxClass dialog (this, -1, "Define search parameters");

  wxCheckBox *SearchtitlesPtr = dialog.GetSearchtitles ();
  wxCheckBox *SearchhintsPtr = dialog.GetSearchhint ();
  wxCheckBox *SearchtextPtr = dialog.GetSearchdescriptive ();
  wxCheckBox *SearchallPtr = dialog.GetSearchall ();
  wxCheckBox *SearchpsacPtr = dialog.GetSearchpsac ();
  wxCheckBox *SearchsdpPtr = dialog.GetSearchsdp ();
  wxCheckBox *SearchsvpPtr = dialog.GetSearchsvp ();
  wxCheckBox *SearchscmpPtr = dialog.GetSearchscmp ();
  wxCheckBox *SearchsqapPtr = dialog.GetSearchsqap ();
  wxCheckBox *SearchsrsPtr = dialog.GetSearchsrs ();
  wxCheckBox *SearchsdsPtr = dialog.GetSearchsds ();
  wxCheckBox *SearchscsPtr = dialog.GetSearchscs ();
  wxCheckBox *SearchsrdPtr = dialog.GetSearchsrd ();
  wxCheckBox *SearchsddPtr = dialog.GetSearchsdd ();
  wxCheckBox *SearchsvcpPtr = dialog.GetSearchsvcp ();
  wxCheckBox *SearchseciPtr = dialog.GetSearchseci ();
  wxCheckBox *SearchsciPtr = dialog.GetSearchsci ();
  wxCheckBox *SearchsasPtr = dialog.GetSearchsas ();
  wxTextCtrl *SearchboxPtr = dialog.GetSearchbox ();
  wxBitmapButton *FindagainbuttonPtr = GetFindagainbutton ();

  // Adapt for DO-178B vs. DO-254.
  extern int Do254;
  if (Do254)
    {
      SearchpsacPtr->SetLabel ("PHAC");
      SearchsdpPtr->SetLabel ("HDP");
      SearchsvpPtr->SetLabel ("HVVP");
      SearchscmpPtr->SetLabel ("HCMP");
      SearchsqapPtr->SetLabel ("HPAP");
      SearchsrsPtr->SetLabel ("RS");
      SearchsdsPtr->SetLabel ("HDS");
      SearchscsPtr->SetLabel ("VVS");
      SearchsrdPtr->SetLabel ("HRD");
      SearchsddPtr->SetLabel ("HDRD");
      SearchsvcpPtr->SetLabel ("HTPR");
      SearchseciPtr->SetLabel ("HATC");
      SearchsciPtr->SetLabel ("HARS");
      SearchsasPtr->SetLabel ("HAS");
    }
  else
    {
      SearchpsacPtr->SetLabel ("PSAC");
      SearchsdpPtr->SetLabel ("SDP");
      SearchsvpPtr->SetLabel ("SVP");
      SearchscmpPtr->SetLabel ("SCMP");
      SearchsqapPtr->SetLabel ("SQAP");
      SearchsrsPtr->SetLabel ("SRS");
      SearchsdsPtr->SetLabel ("SDS");
      SearchscsPtr->SetLabel ("SCS");
      SearchsrdPtr->SetLabel ("SRD");
      SearchsddPtr->SetLabel ("SDD");
      SearchsvcpPtr->SetLabel ("SVCP");
      SearchseciPtr->SetLabel ("SECI");
      SearchsciPtr->SetLabel ("SCI");
      SearchsasPtr->SetLabel ("SAS");
    }

  SearchtitlesPtr->SetValue (SearchTitles);
  SearchhintsPtr->SetValue (SearchHints);
  SearchtextPtr->SetValue (SearchText);
  SearchallPtr->SetValue (SearchAll);
  SearchpsacPtr->SetValue (SearchDocs[PSAC]);
  SearchpsacPtr->Enable (!SearchAll);
  SearchsdpPtr->SetValue (SearchDocs[SDP]);
  SearchsdpPtr->Enable (!SearchAll);
  SearchsvpPtr->SetValue (SearchDocs[SVP]);
  SearchsvpPtr->Enable (!SearchAll);
  SearchscmpPtr->SetValue (SearchDocs[SCMP]);
  SearchscmpPtr->Enable (!SearchAll);
  SearchsqapPtr->SetValue (SearchDocs[SQAP]);
  SearchsqapPtr->Enable (!SearchAll);
  SearchsrsPtr->SetValue (SearchDocs[SRS]);
  SearchsrsPtr->Enable (!SearchAll);
  SearchsdsPtr->SetValue (SearchDocs[SDS]);
  SearchsdsPtr->Enable (!SearchAll);
  SearchscsPtr->SetValue (SearchDocs[SCS]);
  SearchscsPtr->Enable (!SearchAll);
  SearchsrdPtr->SetValue (SearchDocs[SRD]);
  SearchsrdPtr->Enable (!SearchAll);
  SearchsddPtr->SetValue (SearchDocs[SDD]);
  SearchsddPtr->Enable (!SearchAll);
  SearchsvcpPtr->SetValue (SearchDocs[SVCP]);
  SearchsvcpPtr->Enable (!SearchAll);
  SearchseciPtr->SetValue (SearchDocs[SECI]);
  SearchseciPtr->Enable (!SearchAll);
  SearchsciPtr->SetValue (SearchDocs[SCI]);
  SearchsciPtr->Enable (!SearchAll);
  SearchsasPtr->SetValue (SearchDocs[SAS]);
  SearchsasPtr->Enable (!SearchAll);
  SearchboxPtr->SetValue (SearchString);
  dialog.Center ();
  if (dialog.ShowModal ())
    FindagainbuttonPtr->Enable (true);
}

//---------------------------------------------------------------------------
// A handler that gets called when the dialog closes.
// We have to make sure that the in-memory XML structure gets updated
// with correct editor-box info, if the editor is closed.

void
XmlEditorDialogClass::OnClose (wxCloseEvent & event)
{
  XmlChange (ItemTree->GetSelection (), RootXml);
  EndModal (0);
}

//---------------------------------------------------------------------------
// This handler is called on resizing the dialog.  It shouldn't be necessary
// (and isn't in Linux), but in Win32 the toolbar stuff over the text
// control doesn't get redrawn properly.
// ... Oh, and it doesn't work anyhow.

void
XmlEditorDialogClass::OnSize (wxSizeEvent & event)
{
  //MarkupChoice->Refresh ();
  //MarkupButton->Refresh ();
  //NoviceButton->Refresh ();
  //ExpertButton->Refresh ();
  //Refresh ();
}

//---------------------------------------------------------------------------
// A handler called when the "Save" button is clicked.

void
XmlEditorDialogClass::SaveButton (wxCommandEvent & event)
{
  // The purpose of the following lines is to force any data in the
  // text window to be saved.  It will basically mimic clicking
  // on the item with the mouse.
  wxTreeEvent DummyEvent;
  ClickTreeBypass = 1;
  ClickTree (DummyEvent);
  // Okay, save it.
  extern int ReadOnly;
  if (ReadOnly)
    {
      wxMessageBox ("This file was opened read-only.\n"
		    "It can only be saved with /File/SaveAs.", "Error",
		    wxCENTRE | wxOK | wxICON_ERROR);
    }
  else if (SaveXmlFile (rFilename, RootXml))
    wxMessageBox ("Could not save the file.", "Error",
		  wxCENTRE | wxOK | wxICON_ERROR);
  else
    {
      DatasetChanged = 0;
      wxMessageBox ("File saved.", "Status",
		    wxCENTRE | wxOK | wxICON_INFORMATION);
    }
}

// Looks as though I must have gotten started on this once upon
// a time, and then given up on it.  Hmmm.  It doesn't bode well
// for ClickBuildButton, which I just added below.
void
XmlEditorDialogClass::BuildButtonClick (wxCommandEvent & event)
{

}

void XmlEditorDialogClass::ClickBuildButton( wxCommandEvent &event )
{
  extern int BuildTheDocuments (void);
  BuildTheDocuments ();
}

//---------------------------------------------------------------------------
// A handler called when the "Delete" button is clicked.

void
XmlEditorDialogClass::DeleteItem (wxCommandEvent & event)
{
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Record = FindXml (RootXml, Selected);
  struct XmlStruct **Pred = FindPredecessor (RootXml, Record);
  if (Pred == NULL)
    return;
  if (0 == strcmp (Record->Type, "TBD") ||
      wxYES ==
      wxMessageBox ("Are you sure you want to delete this data item\n"
		    "and all of its sub-items (if any)?  If this is\n"
		    "an item originally present in the framework\n"
		    "(in other words, an item not added by you),\n"
		    "then removing it could introduce incompatibilities\n"
		    "with existing project datasets.", "Confirm",
		    wxCENTRE | wxYES_NO | wxICON_QUESTION))
    {
      LastSelected = INVALID_TREE_ITEM;
      XmlChange (Selected, RootXml);
      *Pred = Record->Next;
      if (Record->Children != NULL)
	ClearXml (Record->Children);
      free (Record->Name);
      free (Record->Type);
      free (Record->Hint);
      free (Record->Text);
      free (Record);
      ItemTree->Delete (Selected);
      LastSelected = INVALID_TREE_ITEM;
    }
}

//---------------------------------------------------------------------------
// A handler called when the "Add sub-item" button is clicked.

static const char DefaultHint[] =
  "This is a newly-added item, not present in the original framework.  "
  "Don\'t forget to change the title of the data item and the hint to something "
  "meaningful.  You should probably keep the value in the text-editor (i.e., \"TBD\" "
  "is a good default value).  The document-assignments, applicability, and control-"
  "category settings have been duplicated from the data item above, and should "
  "be changed if they\'re not what\'s needed.";
static const char CustomHint[] =
  "This is a CUSTOM item, not present in the framework.";  

void
XmlEditorDialogClass::AddSubItem (wxCommandEvent & event)
{
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Record = FindXml (RootXml, Selected);
  if (Record == NULL)
    return;
  if (HintEditable || TemporaryHintEditable)
    {
      struct XmlStruct *NewRecord;
      NewRecord = (struct XmlStruct *) malloc (sizeof (struct XmlStruct));
      if (NewRecord == NULL)
	{
	OutOfMemory:
	  wxMessageBox ("Out of memory.\nCannot add further items.\n"
			"Please save and exit.", "Fatal error",
			wxCENTRE | wxOK | wxICON_ERROR);
	  return;
	}
      LastSelected = Record->Id;
      XmlChange (Record->Id, RootXml);
      *NewRecord = *Record;
      NewRecord->Parent = Record;
      NewRecord->Next = Record->Children;
      NewRecord->Children = NULL;
      Record->Children = NewRecord;
      NewRecord->Name = (char *) malloc (4);
      if (NewRecord->Name == NULL)
	goto OutOfMemory;
      strcpy (NewRecord->Name, "TBD");
      NewRecord->Type = (char *) malloc (4);
      if (NewRecord->Type == NULL)
	goto OutOfMemory;
      strcpy (NewRecord->Type, "TBD");
      NewRecord->Text = (char *) malloc (4);
      if (NewRecord->Text == NULL)
	goto OutOfMemory;
      strcpy (NewRecord->Text, "TBD");
      if (TemporaryHintEditable)
        {
	  int i;
	  for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
	    NewRecord->Applicability[i] = CC_DISCRETION | SATISFIED_DISCRETION;
	}
      NewRecord->Hint = (char *) malloc (sizeof (DefaultHint));
      if (NewRecord->Hint == NULL)
	goto OutOfMemory;
      if (TemporaryHintEditable)
        strcpy (NewRecord->Hint, CustomHint);
      else 
        strcpy (NewRecord->Hint, DefaultHint);
      NewRecord->Color = (wxColour *) wxBLACK;
      NewRecord->Id = ItemTree->PrependItem (Record->Id, "TBD");
      ItemText->SetValue (NewRecord->Text);
      HintBox->SetValue (NewRecord->Hint);
      DatasetNewlyChanged = 1;
      LastSelected = NewRecord->Id;
      XmlChange (NewRecord->Id, RootXml);
      ItemTree->Expand (Record->Id);
      ItemTree->SelectItem (NewRecord->Id);
      if (HintEditable)
	{
	  Add->Enable (TRUE);
	  AddSub->Enable (TRUE);
	}
      else
	{
	  Add->Enable (NULL != AllowableAddSiblings (NewRecord));
	  AddSub->Enable (NULL != AllowableAddSubs (NewRecord));
	}
      Del->Enable (TRUE);
    }
  else
    {
      struct XmlStruct *NewRecord =
	CreateTreeCopy (AllowableAddSubs (Record));
      if (NewRecord == NULL)
	{
	  wxMessageBox ("Sorry, cannot create new item.", "Fatal error",
			wxCENTRE | wxOK | wxICON_ERROR);
	  return;
	}
      LastSelected = Record->Id;
      XmlChange (Record->Id, RootXml);
      NewRecord->Parent = Record;
      NewRecord->Next = Record->Children;
      //NewRecord->Children = NULL;
      Record->Children = NewRecord;
      NewRecord->Id = ItemTree->PrependItem (Record->Id, NewRecord->Name);
      FillSubTree (NewRecord);
      ItemText->SetValue (NewRecord->Text);
      HintBox->SetValue (NewRecord->Hint);
      DatasetNewlyChanged = 1;
      LastSelected = NewRecord->Id;
      XmlChange (NewRecord->Id, RootXml);
      ItemTree->Expand (Record->Id);
      ItemTree->SelectItem (NewRecord->Id);
      FrameworkSet (NewRecord);
      if (HintEditable)
	{
	  Add->Enable (TRUE);
	  AddSub->Enable (TRUE);
	}
      else
	{
	  Add->Enable (NULL != AllowableAddSiblings (NewRecord));
	  AddSub->Enable (NULL != AllowableAddSubs (NewRecord));
	}
      Del->Enable (TRUE);
    }
}

//---------------------------------------------------------------------------
// A handler called when the "Add sibling item" button is clicked.

void
XmlEditorDialogClass::AddItem (wxCommandEvent & event)
{
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Record = FindXml (RootXml, Selected);
  if (Record == NULL || Record == RootXml)
    return;
  if (HintEditable || TemporaryHintEditable)
    {
      struct XmlStruct *NewRecord;
      NewRecord = (struct XmlStruct *) malloc (sizeof (struct XmlStruct));
      if (NewRecord == NULL)
	{
	OutOfMemory:
	  wxMessageBox ("Out of memory.\nCannot add further items.\n"
			"Please save and exit.", "Fatal error",
			wxCENTRE | wxOK | wxICON_ERROR);
	  return;
	}
      LastSelected = Record->Id;
      XmlChange (Record->Id, RootXml);
      *NewRecord = *Record;
      NewRecord->Parent = Record->Parent;
      NewRecord->Next = Record->Next;
      NewRecord->Children = NULL;
      Record->Next = NewRecord;
      NewRecord->Name = (char *) malloc (4);
      if (NewRecord->Name == NULL)
	goto OutOfMemory;
      strcpy (NewRecord->Name, "TBD");
      NewRecord->Type = (char *) malloc (4);
      if (NewRecord->Type == NULL)
	goto OutOfMemory;
      strcpy (NewRecord->Type, "TBD");
      NewRecord->Text = (char *) malloc (4);
      if (NewRecord->Text == NULL)
	goto OutOfMemory;
      strcpy (NewRecord->Text, "TBD");
      if (TemporaryHintEditable)
        {
	  int i;
	  for (i = 0; i < NUM_CRITICALITY_LEVELS; i++)
	    NewRecord->Applicability[i] = CC_DISCRETION | SATISFIED_DISCRETION;
	}
      NewRecord->Hint = (char *) malloc (sizeof (DefaultHint));
      if (NewRecord->Hint == NULL)
	goto OutOfMemory;
      if (TemporaryHintEditable)
        strcpy (NewRecord->Hint, CustomHint);
      else 
        strcpy (NewRecord->Hint, DefaultHint);
      NewRecord->Color = (wxColour *) wxBLACK;
      NewRecord->Id =
	ItemTree->InsertItem (NewRecord->Parent->Id, Record->Id, "TBD");
      ItemText->SetValue (NewRecord->Text);
      HintBox->SetValue (NewRecord->Hint);
      DatasetNewlyChanged = 1;
      LastSelected = NewRecord->Id;
      XmlChange (NewRecord->Id, RootXml);
      ItemTree->SelectItem (NewRecord->Id);
      FrameworkSet (NewRecord);
      if (HintEditable)
	{
	  Add->Enable (TRUE);
	  AddSub->Enable (TRUE);
	}
      else
	{
	  Add->Enable (NULL != AllowableAddSiblings (NewRecord));
	  AddSub->Enable (NULL != AllowableAddSubs (NewRecord));
	}
      Del->Enable (TRUE);
    }
  else
    {
      struct XmlStruct *NewRecord =
	CreateTreeCopy (AllowableAddSiblings (Record));
      if (NewRecord == NULL)
	{
	  wxMessageBox ("Sorry, cannot create the new item.", "Fatal error",
			wxCENTRE | wxOK | wxICON_ERROR);
	  return;
	}
      LastSelected = Record->Id;
      XmlChange (Record->Id, RootXml);
      NewRecord->Parent = Record->Parent;
      NewRecord->Next = Record->Next;
      Record->Next = NewRecord;
      NewRecord->Id =
	ItemTree->InsertItem (NewRecord->Parent->Id, Record->Id,
			      NewRecord->Name);
      FillSubTree (NewRecord);
      ItemText->SetValue (NewRecord->Text);
      HintBox->SetValue (NewRecord->Hint);
      DatasetNewlyChanged = 1;
      LastSelected = NewRecord->Id;
      XmlChange (NewRecord->Id, RootXml);
      ItemTree->SelectItem (NewRecord->Id);
      if (HintEditable)
	{
	  Add->Enable (TRUE);
	  AddSub->Enable (TRUE);
	}
      else
	{
	  Add->Enable (NULL != AllowableAddSiblings (NewRecord));
	  AddSub->Enable (NULL != AllowableAddSubs (NewRecord));
	}
      Del->Enable (TRUE);
    }
}

//---------------------------------------------------------------------------
// Given a pointer to a record in the template structure (or to any other
// XmlStruct record), copies the record and its entire subtree to memory,
// allocated space as required.  Returns a pointer to the copy of the subtree,
// or NULL on failure.  Operation is recursive.  On failure, all allocated
// memory is deallocated.

struct XmlStruct *
XmlEditorDialogClass::CreateTreeCopy (struct XmlStruct *Source)
{
  struct XmlStruct *Record, *SourceChild, *Child;
  struct XmlStruct *LastChild = NULL;
  if (Source == NULL)
    return (NULL);
  // Allocate the new record, and duplicate the source record
  Record = (struct XmlStruct *) malloc (sizeof (struct XmlStruct));
  if (Record == NULL)
    return (NULL);
  *Record = *Source;
  Record->Name = (char *) malloc (strlen (Source->Name) + 1);
  if (Record->Name == NULL)
    {
      free (Record);
      return (NULL);
    }
  strcpy (Record->Name, Source->Name);
  Record->Type = (char *) malloc (strlen (Source->Type) + 1);
  if (Record->Type == NULL)
    {
      free (Record->Name);
      free (Record);
      return (NULL);
    }
  strcpy (Record->Type, Source->Type);
  Record->Hint = (char *) malloc (strlen (Source->Hint) + 1);
  if (Record->Hint == NULL)
    {
      free (Record->Type);
      free (Record->Name);
      free (Record);
      return (NULL);
    }
  strcpy (Record->Hint, Source->Hint);
  Record->Text = (char *) malloc (strlen (Source->Text) + 1);
  if (Record->Text == NULL)
    {
      free (Record->Hint);
      free (Record->Type);
      free (Record->Name);
      free (Record);
      return (NULL);
    }
  strcpy (Record->Text, Source->Text);
  Record->Parent = NULL;
  Record->Next = NULL;
  Record->Children = NULL;
  // Loop on the children of the source record.
  for (SourceChild = Source->Children; SourceChild != NULL;
       SourceChild = SourceChild->Next)
    {
      Child = CreateTreeCopy (SourceChild);
      if (Child == NULL)
	{
	  ClearXml (Record);
	  free (Record->Text);
	  free (Record->Hint);
	  free (Record->Type);
	  free (Record->Name);
	  free (Record);
	  return (NULL);
	}
      Child->Parent = Record;
      if (Record->Children == NULL)
	Record->Children = Child;
      else
        {
	  if (LastChild != NULL)
	    LastChild->Next = Child;
	}
      LastChild = Child;
    }
  // All done!
  return (Record);
}

//---------------------------------------------------------------------------
// Given a pointer to an XmlStruct record which has just been added to the 
// ItemTree, adds the sub-tree to the ItemTree as well.  This function
// operates recursively.

void
XmlEditorDialogClass::FillSubTree (struct XmlStruct *Xml)
{
  struct XmlStruct *Next, *Last = NULL;
  for (Next = Xml->Children; Next != NULL; Next = Next->Next)
    {
      if (Next == Xml->Children)
	Next->Id = ItemTree->PrependItem (Xml->Id, Next->Name);
      else
	Next->Id =
	  ItemTree->InsertItem (Next->Parent->Id, Last->Id, Next->Name);
      if (Next->Children != NULL)
	FillSubTree (Next);
      Last = Next;
    }
}

//---------------------------------------------------------------------------
// Handlers called when the various control-category buttons are clicked.

void
XmlEditorDialogClass::ClickCcE (wxCommandEvent & event)
{
  CcChange (RootXml, DO178B_LEVEL_E, CcE);
}

void
XmlEditorDialogClass::ClickCcD (wxCommandEvent & event)
{
  CcChange (RootXml, DO178B_LEVEL_D, CcD);
}

void
XmlEditorDialogClass::ClickCcC (wxCommandEvent & event)
{
  CcChange (RootXml, DO178B_LEVEL_C, CcC);
}

void
XmlEditorDialogClass::ClickCcB (wxCommandEvent & event)
{
  CcChange (RootXml, DO178B_LEVEL_B, CcB);
}

void
XmlEditorDialogClass::ClickCcA (wxCommandEvent & event)
{
  CcChange (RootXml, DO178B_LEVEL_A, CcA);
}

//---------------------------------------------------------------------------
// Handlers called when the various "satisfaction" buttons are clicked.

void
XmlEditorDialogClass::ClickSatE (wxCommandEvent & event)
{
  SatChange (RootXml, DO178B_LEVEL_E, BitmapSatE);
}

void
XmlEditorDialogClass::ClickSatD (wxCommandEvent & event)
{
  SatChange (RootXml, DO178B_LEVEL_D, BitmapSatD);
}

void
XmlEditorDialogClass::ClickSatC (wxCommandEvent & event)
{
  SatChange (RootXml, DO178B_LEVEL_C, BitmapSatC);
}

void
XmlEditorDialogClass::ClickSatB (wxCommandEvent & event)
{
  SatChange (RootXml, DO178B_LEVEL_B, BitmapSatB);
}

void
XmlEditorDialogClass::ClickSatA (wxCommandEvent & event)
{
  SatChange (RootXml, DO178B_LEVEL_A, BitmapSatA);
}

//---------------------------------------------------------------------------
// Handlers that get called when document-type radio-buttons are clicked.

void
XmlEditorDialogClass::SqarClick (wxCommandEvent & event)
{

}

void
XmlEditorDialogClass::ScmrClick (wxCommandEvent & event)
{

}

void
XmlEditorDialogClass::SvrClick (wxCommandEvent & event)
{

}

void
XmlEditorDialogClass::SasClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SAS, SasCheck);
}

void
XmlEditorDialogClass::SciClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SCI, SciCheck);
}

void
XmlEditorDialogClass::SeciClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SECI, SeciCheck);
}

void
XmlEditorDialogClass::SvcpClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SVCP, SvcpCheck);
}

void
XmlEditorDialogClass::DdClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, DD, DdCheck);
}

void
XmlEditorDialogClass::SrdClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SRD, SrdCheck);
}

void
XmlEditorDialogClass::ScsClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SCS, ScsCheck);
}

void
XmlEditorDialogClass::SdsClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SDS, SdsCheck);
}

void
XmlEditorDialogClass::SrsClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SRS, SrsCheck);
}

void
XmlEditorDialogClass::SqapClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SQAP, SqapCheck);
}

void
XmlEditorDialogClass::ScmpClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SCMP, ScmpCheck);
}

void
XmlEditorDialogClass::SvpClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SVP, SvpCheck);
}

void
XmlEditorDialogClass::SdpClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, SDP, SdpCheck);
}

void
XmlEditorDialogClass::PsacClick (wxCommandEvent & event)
{
  DocumentChange (RootXml, PSAC, PsacCheck);
}

//---------------------------------------------------------------------------
// A handler called when outline is clicked.

void
XmlEditorDialogClass::ClickTree (wxTreeEvent & event)
{
  int Bypassing = ClickTreeBypass;
  wxTreeItemId Selected;
  // Default editor button setup.
  Add->Enable (HintEditable);
  AddSub->Enable (HintEditable);
  AddCustom->Enable (TRUE);
  AddSubCustom->Enable (TRUE);
  Del->Enable (HintEditable);
  if (ClickTreeBypass)
    {
      Selected = ItemTree->GetSelection ();
      ClickTreeBypass = 0;
    }
  else
    {
      Selected = event.GetItem ();
      if (Selected == RootXml->Id)
	{
	  Add->Enable (FALSE);
	  Del->Enable (FALSE);
	}
    }
  if (Bypassing || Selected != LastSelected)
    {
      if (Bypassing)
	LastSelected = Selected;
      // Here, need to write whatever was in the editor box to the OLD tree item.
      /*
         The rationale for the funky stuff below is probably bogus,
         because I now think the problems I experienced were due to 
         inadequate buffer sizes in XmlChange blowing the stack.  
         However, here's the possibly-bogus explanation:

         I don't know why (this is the only time I've seen it)
         but the most-significant word of "this" gets overwritten 
         on return from XmlChange at this point.  (It's perfectly okay
         INSIDE of XmlChange.)  Perhaps it's a PPC thing.  It means, though,
         that we have to capture "this" before using XmlChange, and then to
         explicitly use "that" throughout the remainder of the function, 
         rather than supposing that the references to class members will
         work properly.  Using "that->" everywhere is not only tedious and
         confusing, but also prone to error (from being forgotten), so I've
         split the ClickTree function into two parts:  ClickTree and ClickCleanup.
       */
      XmlEditorDialogClass *that = this;
      if (LastSelected != INVALID_TREE_ITEM)
	XmlChange (LastSelected, RootXml);

      // Now set up the display for the new item.
      that->LastSelected = Selected;
      XmlStruct *Record = FindXml (that->RootXml, Selected);
      that->ClickCleanup (Record);
    }
}

// ... continuation of ClickTree, to account for the "this"-corruption error.
// (Later:  I now believe there's no such bug, but the code is okay anyhow.)
// Basically, just sets up the on-screen display for whatever the newly
// selected data record contains.

void
XmlEditorDialogClass::ClickCleanup (struct XmlStruct *Record)
{
  FrameworkSet (Record);
  if (Record != NULL)
    {
      // Hint.
      if (Record->Hint != NULL)
	HintBox->SetValue (Record->Hint);
      else
	HintBox->SetValue ("");
      // Editor text.
      if (Record->Text != NULL)
	ItemText->SetValue (Record->Text);
      else
	ItemText->SetValue ("");
      // Set up the Add/Del/... buttons.
      if (Record->Parent == NULL)
	{
	  Add->Enable (FALSE);
	  AddSub->Enable (NULL != AllowableAddSubs (Record));
	  Del->Enable (FALSE);
	}
      else if (HintEditable)
	{
	  Add->Enable (TRUE);
	  AddSub->Enable (TRUE);
	  Del->Enable (TRUE);
	}
      else
	{
	  Add->Enable (NULL != AllowableAddSiblings (Record));
	  AddSub->Enable (NULL != AllowableAddSubs (Record));
	  Del->Enable (TRUE);
	}
      ItemText->Refresh ();
      HintBox->Refresh ();
    }
  else
    {
      ItemText->SetValue ("");
      HintBox->SetLabel ("");
      ItemText->Refresh ();
      HintBox->Refresh ();
    }
}

//---------------------------------------------------------------------------
// An auxiliary function (not an event-handler, though used by event handlers)
// that updates the in-memory hierarchical data structure, given changes in
// the displayed outline.

#define TYPE_FORMAT \
   "In addition to the visible title, which has been changed,\n" \
   "there is also an underlying invisible title, determining\n" \
   "what types of data items can legally be inserted at this position.\n" \
   "If this is a NEW data item, the two titles should be the same.\n" \
   "But if you are merely changing the title of an OLD data item,\n" \
   "changing the underlying title could make your old datasets\n" \
   "incompatible!  Should I change the underlying title?\n" \
   "    Underlying title was:  \"%s\"\n" \
   "    New underlying title:  \"%s\""

void
XmlEditorDialogClass::XmlChange (wxTreeItemId Selected, struct XmlStruct *Xml)
{
  XmlStruct *Record = FindXml (Xml, Selected);
  if (Record == NULL)
    return;

  // Make sure all of the fields we need to check have been allocated.
  if (Record->Text == NULL)
    {
      Record->Text = (char *) malloc (1 + strlen (""));
      if (Record->Text == NULL)
	{
	OutOfMemory:
	  wxMessageBox ("Out of memory.\nCannot retain further changes\n"
			"Please save and exit.", "Fatal error",
			wxCENTRE | wxOK | wxICON_ERROR);
	  return;
	}
      strcpy (Record->Text, "");
    }
  if (Record->Hint == NULL)
    {
      Record->Hint = (char *) malloc (1 + strlen (""));
      if (Record->Hint == NULL)
	goto OutOfMemory;
      strcpy (Record->Hint, "");
    }

  // Save any changes from the editor window.
  wxString Text = ItemText->GetValue ();
  if (0 != strcmp (Record->Text, Text))
    {
      DatasetNewlyChanged = 1;
      free (Record->Text);
      Record->Text = (char *) malloc (Text.Len () + 1);
      if (Record->Text == NULL)
	goto OutOfMemory;
      strcpy (Record->Text, Text);
    }

  // Save any changes to the data-item name.
  Text = ItemTree->GetItemText (Selected);
  if (0 != strcmp (Record->Name, Text))
    {
      DatasetNewlyChanged = 1;
      free (Record->Name);
      Record->Name = (char *) malloc (Text.Len () + 1);
      if (Record->Name == NULL)
	goto OutOfMemory;
      strcpy (Record->Name, Text);
    }

  // If we are in framework-modification mode, the same change to the name
  // must also apply to the type.
  if (HintEditable)
    {
      void NameToTag (const char *Name, char *Tag);
      char Tag[513], Warning[strlen (TYPE_FORMAT) + sizeof (Tag) / 3];
      Text = ItemTree->GetItemText (Selected);
      // The following check is to prevent buffer overflows in the 
      // prompt and the tag.  The worst possible expansion in a tag
      // is if EVERY character is converted to something like &#n;
      // which would be a maximum expansion of 6-to-1.
      if (Text.Len () < (sizeof (Tag) - 1) / 6)
	{
	  NameToTag (Text, Tag);
	  if (0 != strcmp (Record->Type, Tag))
	    {
	      sprintf (Warning, TYPE_FORMAT, Record->Type, Tag);
	      if (0 == strcmp (Record->Type, "")
		  || 0 == strcmp (Record->Type, "TBD")
		  || wxYES == wxMessageBox (Warning, "Compatibility danger!",
					    wxYES_NO | wxCENTRE |
					    wxICON_EXCLAMATION))
		{
		  DatasetNewlyChanged = 1;
		  free (Record->Type);
		  Record->Type = (char *) malloc (strlen (Tag) + 1);
		  if (Record->Type == NULL)
		    goto OutOfMemory;
		  strcpy (Record->Type, Tag);
		}
	    }
	}
    }

  // Save any changes to the hint window.
  Text = HintBox->GetValue ();
  if (0 != strcmp (Record->Hint, Text))
    {
      DatasetNewlyChanged = 1;
      free (Record->Hint);
      Record->Hint = (char *) malloc (Text.Len () + 1);
      if (Record->Hint == NULL)
	goto OutOfMemory;
      strcpy (Record->Hint, Text);
    }

  // Note that any framework changes (the relevant-document list, 
  // applicability settings, the control category settings) will 
  // already have been made automatically, since they are associated
  // with events attached to check-boxes and buttons.  

  // Visually mark this item, and all ancestors, as changed.
  if (DatasetNewlyChanged)
    {
      DatasetChanged = 1;
      DatasetNewlyChanged = 0;
      XmlStruct *Parent;
      for (Parent = Record; Parent != NULL; Parent = Parent->Parent)
	{
	  Parent->Color = CHANGE_COLOR;
	  ItemTree->SetItemTextColour (Parent->Id, *CHANGE_COLOR);
	}
    }
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) that can
// locate the predecessor (either the prior sibling or else the parent)
// of any given data item.  That predecessor, and no other data item, 
// contains a pointer back to the data item we are finding the predecessor
// of.  This function returns a pointer to that pointer (which will be
// either the Next or Children field of the predecessor data item).
// If this fails, which can happen only if we started from the root element,
// NULL is returned.

struct XmlStruct **
XmlEditorDialogClass::FindPredecessor (struct XmlStruct *Xml,
				       struct XmlStruct *Record)
{
  struct XmlStruct *Pred;
  if (Record == NULL || Record == Xml)
    return (NULL);
  // We can always find the predecessor by moving back to the
  // Parent, and then looping on the Parent's Children.
  // We have to be a little careful with this algorithm, since
  // we don't ever want to be fooled into returning &Pred.
  if (Record->Parent->Children == Record)
    {
      // In this case the record was the FIRST child of its parent.
      return (&(Record->Parent->Children));
    }
  for (Pred = Record->Parent->Children; Pred->Next != Record;
       Pred = Pred->Next);
  return (&(Pred->Next));
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) that 
// sets all of the labels for a bitmap button to the same bitmap.

void
XmlEditorDialogClass::ButtonBitmap (wxBitmapButton * Button, wxBitmap Bitmap)
{
  Button->SetBitmapLabel (Bitmap);
  Button->SetBitmapFocus (Bitmap);
  Button->SetBitmapDisabled (Bitmap);
  Button->SetBitmapSelected (Bitmap);
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) this 
// changes the in-memory hierarchical data structure according to an 
// action on a CC button.

void
XmlEditorDialogClass::CcChange (struct XmlStruct *Xml, int Level,
				wxBitmapButton * Button)
{
  int i;
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Record = FindXml (Xml, Selected);
  if (Record == NULL)
    return;

  DatasetNewlyChanged = 1;
  i = Record->Applicability[Level] & CC_MASK;
  i += 16;
  if (i != CC_DISCRETION && i != CC_1 && i != CC_2)
    i = 0;
  Record->Applicability[Level] =
    i | (Record->Applicability[Level] & ~CC_MASK);
  ButtonBitmap (Button, (CcFunc (i >> 4)));
  XmlChange (Selected, Xml);
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) that 
// changes the in-memory hierarchical data structure according to an 
// action on a Sat button.

void
XmlEditorDialogClass::SatChange (struct XmlStruct *Xml, int Level,
				 wxBitmapButton * Button)
{
  int i;
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Record = FindXml (Xml, Selected);
  if (Record == NULL)
    return;

  DatasetNewlyChanged = 1;
  i = Record->Applicability[Level] & SATISFIED_MASK;
  i++;
  if (i != SATISFIED_DISCRETION && i != SATISFIED_YES
      && i != SATISFIED_INDEPENDENCE)
    i = 0;
  Record->Applicability[Level] =
    i | (Record->Applicability[Level] & ~SATISFIED_MASK);
  ButtonBitmap (Button, SatisfactionFunc (i));
  XmlChange (Selected, Xml);
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) that 
// looks at one of the document check-boxes, and updates the in-memory
// hierarchical data structure accordingly.

void
XmlEditorDialogClass::DocumentChange (struct XmlStruct *Xml, int DocNum,
				      wxCheckBox * Box)
{
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Record = FindXml (Xml, Selected);
  if (Record == NULL)
    return;

  DatasetNewlyChanged = 1;
  Record->Documents[DocNum] = Box->GetValue ();
  XmlChange (Selected, Xml);
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) that 
// descends recursively through all levels of the in-memory hierarchical data
// structure, to set the "by section" document dependencies.
// On first entry, all of the section boxes must be unchecked.
// In other words, it detects which documents are affected by a given data
// item OR any of its descendents.

void
XmlEditorDialogClass::SetSectionDocs (struct XmlStruct *Record)
{
  PsacSection->SetValue (Record->Documents[PSAC] || PsacSection->GetValue ());
  SdpSection->SetValue (Record->Documents[SDP] || SdpSection->GetValue ());
  SvpSection->SetValue (Record->Documents[SVP] || SvpSection->GetValue ());
  ScmpSection->SetValue (Record->Documents[SCMP] || ScmpSection->GetValue ());
  SqapSection->SetValue (Record->Documents[SQAP] || SqapSection->GetValue ());
  SrsSection->SetValue (Record->Documents[SRS] || SrsSection->GetValue ());
  SdsSection->SetValue (Record->Documents[SDS] || SdsSection->GetValue ());
  ScsSection->SetValue (Record->Documents[SCS] || ScsSection->GetValue ());
  SrdSection->SetValue (Record->Documents[SRD] || SrdSection->GetValue ());
  DdSection->SetValue (Record->Documents[DD] || DdSection->GetValue ());
  SvcpSection->SetValue (Record->Documents[SVCP] || SvcpSection->GetValue ());
  SeciSection->SetValue (Record->Documents[SECI] || SeciSection->GetValue ());
  SciSection->SetValue (Record->Documents[SCI] || SciSection->GetValue ());
  SasSection->SetValue (Record->Documents[SAS] || SasSection->GetValue ());
  for (Record = Record->Children; Record != NULL; Record = Record->Next)
    SetSectionDocs (Record);
}

//---------------------------------------------------------------------------
// An auxiliary function (not a handler, though used by handlers) that can 
// set up the CC, Sat, and Document boxes appropriately for any given data
// item.

void
XmlEditorDialogClass::FrameworkSet (struct XmlStruct *Record)
{
  PsacSection->SetValue (FALSE);
  SdpSection->SetValue (FALSE);
  SvpSection->SetValue (FALSE);
  ScmpSection->SetValue (FALSE);
  SqapSection->SetValue (FALSE);
  SrsSection->SetValue (FALSE);
  SdsSection->SetValue (FALSE);
  ScsSection->SetValue (FALSE);
  SrdSection->SetValue (FALSE);
  DdSection->SetValue (FALSE);
  SvcpSection->SetValue (FALSE);
  SeciSection->SetValue (FALSE);
  SciSection->SetValue (FALSE);
  SasSection->SetValue (FALSE);
  if (Record == NULL)
    {
      PsacCheck->SetValue (FALSE);
      SdpCheck->SetValue (FALSE);
      SvpCheck->SetValue (FALSE);
      ScmpCheck->SetValue (FALSE);
      SqapCheck->SetValue (FALSE);
      SrsCheck->SetValue (FALSE);
      SdsCheck->SetValue (FALSE);
      ScsCheck->SetValue (FALSE);
      SrdCheck->SetValue (FALSE);
      DdCheck->SetValue (FALSE);
      SvcpCheck->SetValue (FALSE);
      SeciCheck->SetValue (FALSE);
      SciCheck->SetValue (FALSE);
      SasCheck->SetValue (FALSE);
      ButtonBitmap (BitmapSatA, SatisfactionFunc (0));
      ButtonBitmap (BitmapSatB, SatisfactionFunc (0));
      ButtonBitmap (BitmapSatC, SatisfactionFunc (0));
      ButtonBitmap (BitmapSatD, SatisfactionFunc (0));
      ButtonBitmap (BitmapSatE, SatisfactionFunc (0));
      ButtonBitmap (CcA, CcFunc (0));
      ButtonBitmap (CcB, CcFunc (0));
      ButtonBitmap (CcC, CcFunc (0));
      ButtonBitmap (CcD, CcFunc (0));
      ButtonBitmap (CcE, CcFunc (0));
    }
  else
    {
      PsacCheck->SetValue (Record->Documents[PSAC]);
      SdpCheck->SetValue (Record->Documents[SDP]);
      SvpCheck->SetValue (Record->Documents[SVP]);
      ScmpCheck->SetValue (Record->Documents[SCMP]);
      SqapCheck->SetValue (Record->Documents[SQAP]);
      SrsCheck->SetValue (Record->Documents[SRS]);
      SdsCheck->SetValue (Record->Documents[SDS]);
      ScsCheck->SetValue (Record->Documents[SCS]);
      SrdCheck->SetValue (Record->Documents[SRD]);
      DdCheck->SetValue (Record->Documents[DD]);
      SvcpCheck->SetValue (Record->Documents[SVCP]);
      SeciCheck->SetValue (Record->Documents[SECI]);
      SciCheck->SetValue (Record->Documents[SCI]);
      SasCheck->SetValue (Record->Documents[SAS]);
      SetSectionDocs (Record);
      ButtonBitmap (BitmapSatA,
		    SatisfactionFunc (Record->Applicability[DO178B_LEVEL_A] &
				      SATISFIED_MASK));
      ButtonBitmap (BitmapSatB,
		    SatisfactionFunc (Record->Applicability[DO178B_LEVEL_B] &
				      SATISFIED_MASK));
      ButtonBitmap (BitmapSatC,
		    SatisfactionFunc (Record->Applicability[DO178B_LEVEL_C] &
				      SATISFIED_MASK));
      ButtonBitmap (BitmapSatD,
		    SatisfactionFunc (Record->Applicability[DO178B_LEVEL_D] &
				      SATISFIED_MASK));
      ButtonBitmap (BitmapSatE,
		    SatisfactionFunc (Record->Applicability[DO178B_LEVEL_E] &
				      SATISFIED_MASK));
      ButtonBitmap (CcA,
		    CcFunc ((Record->Applicability[DO178B_LEVEL_A] & CC_MASK)
			    >> 4));
      ButtonBitmap (CcB,
		    CcFunc ((Record->Applicability[DO178B_LEVEL_B] & CC_MASK)
			    >> 4));
      ButtonBitmap (CcC,
		    CcFunc ((Record->Applicability[DO178B_LEVEL_C] & CC_MASK)
			    >> 4));
      ButtonBitmap (CcD,
		    CcFunc ((Record->Applicability[DO178B_LEVEL_D] & CC_MASK)
			    >> 4));
      ButtonBitmap (CcE,
		    CcFunc ((Record->Applicability[DO178B_LEVEL_E] & CC_MASK)
			    >> 4));
    }
  // The following stuff shouldn't be necessary -- and isn't in Linux -- but in Win32
  // the buttons don't get refreshed for some reason.  
  BitmapSatA->Refresh ();
  BitmapSatB->Refresh ();
  BitmapSatC->Refresh ();
  BitmapSatD->Refresh ();
  BitmapSatE->Refresh ();
  CcA->Refresh ();
  CcB->Refresh ();
  CcC->Refresh ();
  CcD->Refresh ();
  CcE->Refresh ();
}

//--------------------------------------------------------------------------
// Given a pointer to an XmlStruct record, descends through the complete
// data structure and the template structure, comparing records, until it
// finds the template record associated with the specified data record.
// Returns NULL on failure.  The item needs to be found at the correct
// level of the hierarchy, with the correct ancestors, but the ordering
// of the siblings is ignored.

#define MAX_DEPTH 32
struct XmlStruct *
XmlEditorDialogClass::SeekTemplate (struct XmlStruct *Record)
{
  struct XmlStruct *Trace[MAX_DEPTH];
  struct XmlStruct *d;
  int Depth;
  if (Template == NULL || Template->Children == NULL)
    return (NULL);
  // In the Trace array, we store a trail of Record's ancestors up to 
  // the root of the data tree;
  for (Depth = 0, d = Record; d != NULL; d = d->Parent)
    Trace[Depth++] = d;
  // Now descend through the Template, looking for the records corresponding
  // to these ancestors.
  Depth -= 2;			// Trace[Depth] == RootXml->Children.
  for (d = Template->Children; Depth >= 0; Depth--)
    {
      for (; d != NULL; d = d->Next)
	if (!strcmp (d->Type, Trace[Depth]->Type))
	  break;
      if (d == NULL)
	return (NULL);
      if (Depth == 0)
	return (d);
      d = d->Children;
    }
  return (NULL);
}

//--------------------------------------------------------------------------

struct XmlStruct *
XmlEditorDialogClass::AllowableAddSiblings (struct XmlStruct *Record)
{
  // Find the associated template record.
  struct XmlStruct *t = SeekTemplate (Record);
  if (t == NULL)
    return (NULL);
  // A sibling can be added under a variety of circumstances, and these
  // have to be arranged in a priority structure, to insure that certain
  // possibilities aren't locked out permanently when multiple possibilities
  // exist.  Here are the cases, in order of preference:
  //    1.  A record of the SAME TYPE as the template sibling does not follow
  //        the record.  This is actually two subcases:  NO sibling follows
  //        the record, or else a sibling that doesn't match follows.
  //        In either case, the template sibling can be added.
  //    2.  The template sibling is of type "TBD ..." and doesn't match
  //        the sibling of the current record.  Then the template sibling
  //        can be added.  As it happens, this is included in case 1.
  //    3.  The current/template records are of type "TBD ...".  Then the
  //        template item (not its sibling) can be added.
  //    4.  The template sibling is of type "TBD ...".  Then the template
  //        sibling can be added, irrespective of the current sibling.
  if (t->Next != NULL)
    {
      if (Record->Next == NULL)
	return (t->Next);
      if (strcmp (t->Next->Type, Record->Next->Type))
	return (t->Next);
    }
  if (!strncmp (t->Type, "TBD", 3))
    return (t);
  if (t->Next != NULL && !strncmp (t->Next->Type, "TBD", 3))
    return (t->Next);
  return (NULL);
}

//--------------------------------------------------------------------------

struct XmlStruct *
XmlEditorDialogClass::AllowableAddSubs (struct XmlStruct *Record)
{
  // Find the associated template record.
  struct XmlStruct *t = SeekTemplate (Record);
  if (t == NULL)
    return (NULL);
  // The only allowed children which can be added correspond to the 
  // first child of the template record.  This child can be added
  // EITHER if it is missing from the data structure, OR if its
  // type begins with TBD.
  t = t->Children;
  if (t == NULL)
    return (NULL);
  if (Record->Children == NULL)
    return (t);
  if (!strncmp (t->Type, "TBD", 3))
    return (t);
  if (strcmp (t->Type, Record->Children->Type))
    return (t);
  return (NULL);
}

//-------------------------------------------------------------------------

void
XmlEditorDialogClass::ClickInsertGraphic (wxCommandEvent & event)
{
  long from, to, n;
  int Inside;
  wxString s;

  ItemText->GetSelection (&from, &to);
  if (from == to)
    {
      Inside = InsideDocbook (ItemText->GetValue ().Truncate (to));
      s = "\n\n";
      if (!Inside)
	s.Append ("<docbook>");
      s.Append ("<informalfigure><mediaobject>"
		"\n <imageobject><imagedata fileref=\"TBD\" align=\"center\"></imageobject>"
		"\n <caption><para>TBD</para></caption>"
		"\n</mediaobject></informalfigure>");
      if (!Inside)
	s.Append ("</docbook>");
      s.Append ("\n\n");
      n = s.First ("TBD");
      ItemText->SetSelection (to, to);
      ItemText->SetInsertionPoint (to);
      ItemText->WriteText (s);
      ItemText->SetSelection (to + n, to + n + 3);
      DatasetNewlyChanged = 1;
      ItemText->SetFocus ();
    }
}

void XmlEditorDialogClass::InsertMdashClick( wxCommandEvent &event )
{
  long from, to;
  int Inside;
  wxString s;

  ItemText->GetSelection (&from, &to);
  if (from == to)
    {
      Inside = InsideDocbook (ItemText->GetValue ().Truncate (to));
      s = "";
      if (!Inside)
	s.Append ("<docbook>");
      s.Append ("&mdash;");
      if (!Inside)
	s.Append ("</docbook>");
      ItemText->SetSelection (to, to);
      ItemText->SetInsertionPoint (to);
      ItemText->WriteText (s);
      DatasetNewlyChanged = 1;
      ItemText->SetFocus ();
    }
}

void XmlEditorDialogClass::InsertFootnoteClick( wxCommandEvent &event )
{
  long from, to, n;
  int Inside;
  wxString s;

  ItemText->GetSelection (&from, &to);
  if (from == to)
    {
      Inside = InsideDocbook (ItemText->GetValue ().Truncate (to));
      s = "";
      if (!Inside)
	s.Append ("<docbook>");
      s.Append ("<footnote><para>TBD</para></footnote>");
      if (!Inside)
	s.Append ("</docbook>");
      n = s.First ("TBD");
      ItemText->SetSelection (to, to);
      ItemText->SetInsertionPoint (to);
      ItemText->WriteText (s);
      ItemText->SetSelection (to + n, to + n + 3);
      DatasetNewlyChanged = 1;
      ItemText->SetFocus ();
    }
}

void XmlEditorDialogClass::InsertHyperlinkClick( wxCommandEvent &event )
{
  long from, to, n;
  int Inside;
  wxString s;

  ItemText->GetSelection (&from, &to);
  if (from == to)
    {
      Inside = InsideDocbook (ItemText->GetValue ().Truncate (to));
      s = "";
      if (!Inside)
	s.Append ("<docbook>");
      s.Append ("<ulink url=\"http://TBD\"><citetitle>TBD</citetitle></ulink>");
      if (!Inside)
	s.Append ("</docbook>");
      n = s.First ("TBD");
      ItemText->SetSelection (to, to);
      ItemText->SetInsertionPoint (to);
      ItemText->WriteText (s);
      ItemText->SetSelection (to + n, to + n + 3);
      DatasetNewlyChanged = 1;
      ItemText->SetFocus ();
    }
}

enum
{ LIST_BULLETS, LIST_ARABIC, LIST_LOWER_ALPHA,
  LIST_LOWER_ROMAN, LIST_UPPER_ALPHA, LIST_UPPER_ROMAN
};
void
XmlEditorDialogClass::ClickInsertList (wxCommandEvent & event)
{
  long from, to, n, NumRows, Type;
  int Continue = 0, InheritNum = 0;
  int Inside;
  wxString s;
  wxArrayString Types;

  Types.Add ("Bullets");
  Types.Add ("1, 2, 3, 4, ...");
  Types.Add ("a, b, c, d, ...");
  Types.Add ("i, ii, iii, iv, ...");
  Types.Add ("A, B, C, D, ...");
  Types.Add ("I, II, III, IV, ...");

  ItemText->GetSelection (&from, &to);
  if (from != to)
    return;

  s = wxGetSingleChoice ("What type of list should this be?",
			 "List setup wizard", Types);
  n = Types.Index (s);
  if (n == wxNOT_FOUND)
    return;
  Type = n;

  n = wxGetNumberFromUser ("How many rows would you like initially?",
			   "Enter a number (1-200):",
			   "List setup wizard", 10, 1, 200);
  if (n < 1 || n > 200)
    return;
  NumRows = n;

  if (Type != LIST_BULLETS)
    {

      n =
	wxMessageBox
	("Should numeration continue from preceding lists (if any)?",
	 "List setup wizard", wxYES_NO | wxCANCEL | wxICON_QUESTION);
      if (n == wxCANCEL)
	return;
      Continue = (n == wxYES);

      n =
	wxMessageBox ("Should numeration inherit from parent list (if any)?",
		      "List setup wizard",
		      wxYES_NO | wxCANCEL | wxICON_QUESTION);
      if (n == wxCANCEL)
	return;
      InheritNum = (n == wxYES);
    }

  Inside = InsideDocbook (ItemText->GetValue ().Truncate (to));
  s = "\n\n";
  if (!Inside)
    s.Append ("<docbook>");
  switch (Type)
    {
    case LIST_BULLETS:
      s.Append ("<itemizedlist");
      break;
    case LIST_ARABIC:
      s.Append ("\n<orderedlist numeration=\"arabic\"");
      break;
    case LIST_LOWER_ALPHA:
      s.Append ("\n<orderedlist numeration=\"loweralpha\"");
      break;
    case LIST_LOWER_ROMAN:
      s.Append ("\n<orderedlist numeration=\"lowerroman\"");
      break;
    case LIST_UPPER_ALPHA:
      s.Append ("\n<orderedlist numeration=\"upperalpha\"");
      break;
    case LIST_UPPER_ROMAN:
      s.Append ("\n<orderedlist numeration=\"upperroman\"");
      break;
    }
  if (Type != LIST_BULLETS)
    {
      if (Continue)
	s.Append (" continuation=\"continues\"");
      else
	s.Append (" continuation=\"restarts\"");
      if (InheritNum)
	s.Append (" inheritnum=\"inherit\"");
      else
	s.Append (" inheritnum=\"ignore\"");
    }
  s.Append (">");
  for (n = 0; n < NumRows; n++)
    s.Append ("\n <listitem><para>TBD</para></listitem>");
  if (Type == LIST_BULLETS)
    s.Append ("\n</itemizedlist>");
  else
    s.Append ("\n</orderedlist>");
  if (!Inside)
    s.Append ("</docbook>");
  s.Append ("\n\n");
  n = s.First ("TBD");
  ItemText->SetSelection (to, to);
  ItemText->SetInsertionPoint (to);
  ItemText->WriteText (s);
  ItemText->SetSelection (to + n, to + n + 3);
  DatasetNewlyChanged = 1;
  ItemText->SetFocus ();
}

enum
{ WIDTH_DONT_CARE, WIDTH_PROPORTIONAL, WIDTH_IMPERIAL,
  WIDTH_METRIC, WIDTH_POINTS
};

void
XmlEditorDialogClass::ClickInsertTable (wxCommandEvent & event)
{
#define MAX_COLS 12
  long from, to, i, n, NumCols, NumRows, WidthType;
  int Inside;
  int Headings = 1;
  wxString s;
  wxArrayString PossibleWidths;

  PossibleWidths.Add ("Don\'t care");
  PossibleWidths.Add ("Proportional (1.*, 2.*, etc.)");
  PossibleWidths.Add ("Imperial (1in, 2in, etc.)");
  PossibleWidths.Add ("Metric (10mm, 20mm, etc.)");
  PossibleWidths.Add ("Points (72pt, 144pt. etc.)");

  n = wxGetNumberFromUser ("How many columns would you like initially?\n"
			   "You will find this is nasty to change later,\n"
			   "so take extra care to get it correct now!",
			   "Enter a number (1-12):",
			   "List setup wizard", 3, 1, MAX_COLS);
  if (n < 1 || n > MAX_COLS)
    return;
  NumCols = n;



  n = wxGetNumberFromUser ("How many rows would you like initially?",
			   "Enter a number (1-200):",
			   "List setup wizard", 10, 1, 200);
  if (n < 1 || n > 200)
    return;
  NumRows = n;

  s = wxGetSingleChoice ("How would you like to specify the column widths?\n"
			 "Note that there are many more ways of doing this\n"
			 "than we\'re providing choices for.",
			 "List setup wizard", PossibleWidths);
  n = PossibleWidths.Index (s);
  if (n == wxNOT_FOUND)
    return;
  WidthType = n;

  n = wxMessageBox ("Would you like to have column headings?",
		    "List setup wizard",
		    wxYES_NO | wxCANCEL | wxICON_QUESTION);
  if (n == wxCANCEL)
    return;
  Headings = (n == wxYES);

  ItemText->GetSelection (&from, &to);
  if (from == to)
    {
      Inside = InsideDocbook (ItemText->GetValue ().Truncate (to));
      s = "\n\n";
      if (!Inside)
	s.Append ("<docbook>");
      s.Append ("<informaltable frame=\"all\">" "\n<tgroup cols=\"");
      s << NumCols;
      s.Append ("\">");
      for (n = 0; n < NumCols; n++)
	{
	  switch (WidthType)
	    {
	    case WIDTH_DONT_CARE:
	      s.Append ("\n <colspec colwidth=\"1*\">");
	      break;
	    case WIDTH_PROPORTIONAL:
	      s.Append ("\n <colspec colwidth=\"TBD*\">");
	      break;
	    case WIDTH_IMPERIAL:
	      s.Append ("\n <colspec colwidth=\"TBDin\">");
	      break;
	    case WIDTH_METRIC:
	      s.Append ("\n <colspec colwidth=\"TBDmm\">");
	      break;
	    case WIDTH_POINTS:
	      s.Append ("\n <colspec colwidth=\"TBDpt\">");
	      break;
	    }
	}
      if (Headings)
	{
	  s.Append ("\n<thead valign=\"middle\">" "\n <row>");
	  for (n = 0; n < NumCols; n++)
	    s.Append ("\n  <entry>TBD</entry>");
	  s.Append (" </row>" "\n</thead>");
	}
      s.Append ("\n<tbody>\n");
      for (i = 0; i < NumRows; i++)
	{
	  s.Append ("\n <row>");
	  for (n = 0; n < NumCols; n++)
	    s.Append ("\n  <entry>TBD</entry>");
	  s.Append ("\n </row>\n");
	}
      s.Append ("\n</tbody>" "\n</tgroup>" "\n</informaltable>");
      if (!Inside)
	s.Append ("</docbook>");
      s.Append ("\n\n");
      n = s.First ("TBD");
      ItemText->SetSelection (to, to);
      ItemText->SetInsertionPoint (to);
      ItemText->WriteText (s);
      ItemText->SetSelection (to + n, to + n + 3);
      DatasetNewlyChanged = 1;
      ItemText->SetFocus ();
    }
#undef MAX_COLS
}

// This function looks for a tree item of a given doc type, among either
// the children of a given item, the downstream siblings of a given item,
// or the downstream siblings of the parent.  To do a complete search, 
// just do three calls:  First to CHILDREN, next to SIBLINGS, and finally
// to PARENT.  ... Later:  Also, then, you have to check the siblings of
// the grandparents, the great-grandparents, and so on.
// It returns either a pointer to the item, or else NULL.

enum
{ CHILDREN, SIBLINGS, PARENT };

static XmlStruct *
ItemSearchByDocType (struct XmlStruct *Xml, int Index, int Who)
{
  struct XmlStruct *Next, *RetVal = NULL;
  if (Xml == NULL)
    return (NULL);
  if (Who == SIBLINGS)
    Next = Xml->Next;
  else if (Who == CHILDREN)
    Next = Xml->Children;
  else if (Who == PARENT)
    {
      if (Xml->Parent == NULL)
	return (NULL);
      Next = (Xml->Parent)->Next;
    }
  else
    return (NULL);
  for (; Next != NULL; Next = Next->Next)
    {
      if (Next->Documents[Index] == DOCUMENT_YES)
	return (Next);
      RetVal = ItemSearchByDocType (Next, Index, CHILDREN);
      if (RetVal != NULL)
	return (RetVal);
    }
  return (NULL);
}

void
XmlEditorDialogClass::ClickDocNext (wxCommandEvent & event)
{
  int Index;
  struct XmlStruct *RetVal, *Next;

  // Returns the index of the document type we're looking for.
  Index = DocChoice->GetSelection ();
  if (Index == wxNOT_FOUND)
    return;

  // Finds the current item in the database.
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Current = FindXml (RootXml, Selected);
  if (Current == NULL)
    return;

  // Seek the next item of this type.
  RetVal = ItemSearchByDocType (Current, Index, CHILDREN);
  if (RetVal == NULL)
    {
      RetVal = ItemSearchByDocType (Current, Index, SIBLINGS);
      if (RetVal == NULL)
	{
	  for (Next = Current; Next != NULL; Next = Next->Parent)
	    {
	      RetVal = ItemSearchByDocType (Next, Index, PARENT);
	      if (RetVal != NULL)
		break;
	    }
	}
    }

  // Open 'er up!
  if (RetVal != NULL)
    {
      wxTreeEvent DummyEvent;
      for (Next = RetVal; Next != NULL; Next = Next->Parent)
	ItemTree->Expand (Next->Id);
      ItemTree->SelectItem (RetVal->Id);
      ClickTreeBypass = 1;
      ClickTree (DummyEvent);
    }

}

// Similar to the ItemSearchByDocType2 function, except that
// it keeps searching until it hits the Current item.  It returns
// 0 if the endpoint was reached, non-zero otherwise.
int
ItemSearchByDocType2 (struct XmlStruct *Xml, int Index,
		      struct XmlStruct *Current, struct XmlStruct **Found)
{
  struct XmlStruct *Next;
  if (Xml == NULL)
    return (1);
  for (Next = Xml->Children; Next != NULL; Next = Next->Next)
    {
      if (Next == Current)
	return (0);
      if (Next->Documents[Index] == DOCUMENT_YES)
	*Found = Next;
      if (!ItemSearchByDocType2 (Next, Index, Current, Found))
	return (0);
    }
  return (1);
}

void
XmlEditorDialogClass::ClickDocBack (wxCommandEvent & event)
{
  int Index;
  struct XmlStruct *RetVal, *Next;

  // Returns the index of the document type we're looking for.
  Index = DocChoice->GetSelection ();
  if (Index == wxNOT_FOUND)
    return;

  // Finds the current item in the database.
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Current = FindXml (RootXml, Selected);
  if (Current == NULL)
    return;

  // Seek the next item of this type.
  RetVal = NULL;
  ItemSearchByDocType2 (RootXml, Index, Current, &RetVal);

  // Open 'er up!
  if (RetVal != NULL)
    {
      wxTreeEvent DummyEvent;
      for (Next = RetVal; Next != NULL; Next = Next->Parent)
	ItemTree->Expand (Next->Id);
      ItemTree->SelectItem (RetVal->Id);
      ClickTreeBypass = 1;
      ClickTree (DummyEvent);
    }

}

// Like ItemSearchByDocType, but with different constraints,
// suited to FindAgain.  If the match was found in the text,
// then from,to defines a new selection area.  If not, (if,
// for example, the match was in titles or hints), then
// from,to will be returned as 0,0.
static XmlStruct *
ItemSearchByDocType3 (struct XmlStruct *Xml,
		      int Who, 
		      char *SearchString,
		      bool *SearchDocs,
		      bool SearchAll, 
		      bool SearchTitles, 
		      bool SearchHints, 
		      bool SearchText,
		      long *from,
		      long *to)
{
  struct XmlStruct *Next, *RetVal = NULL;
  int Length;
  Length = strlen (SearchString);
  if (Xml == NULL)
    return (NULL);
  if (Who == SIBLINGS)
    Next = Xml->Next;
  else if (Who == CHILDREN)
    Next = Xml->Children;
  else if (Who == PARENT)
    {
      if (Xml->Parent == NULL)
	return (NULL);
      Next = (Xml->Parent)->Next;
    }
  else
    return (NULL);
  for (; Next != NULL; Next = Next->Next)
    {
      int TypeMatch, i;
      char *s;
      
      // Should we even search this type of document?
      TypeMatch = SearchAll;
      if (!TypeMatch)
        {
	  for (i = 0; i < NUM_DOCUMENTS; i++)
	    if (SearchDocs[i] && Next->Documents[i] == DOCUMENT_YES)
	      {
	        TypeMatch = 1;
		break;
	      }
	}
	
	
      // Yes, the document type is okay.  Search the appropriate
      // fields.
      if (TypeMatch)
        { 
		
	  if (SearchText)
	    {
	      for (s = Next->Text; *s; s++)
		if (!strncasecmp (s, SearchString, Length))
		  {
		    *from = s - Next->Text;
		    *to = *from + Length;
		    return (Next);
		  }
	    }      	
	
	  if (SearchHints)
	    {
	      for (s = Next->Hint; *s; s++)
		if (!strncasecmp (s, SearchString, Length))
		  {
		    *from = 0;
		    *to = 0;
		    return (Next);
		  }
	    }
	
	  if (SearchTitles)
	    {
	      for (s = Next->Name; *s; s++)
		if (!strncasecmp (s, SearchString, Length))
		  {
		    *from = 0;
		    *to = 0;
		    return (Next);
		  }
	    }
	
	}
	
      // No match with this item.  Continue to lower levels.	
      RetVal = ItemSearchByDocType3 (Next, 
      				     CHILDREN,
				     SearchString,
				     SearchDocs,
				     SearchAll,
				     SearchTitles,
				     SearchHints,
				     SearchText,
				     from,
				     to);
      if (RetVal != NULL)
	return (RetVal);
    }
  return (NULL);
}

//-----------------------------------------------------------------
// A handler that gets called when the "find again" button is 
// pressed.  It actually does all of the actual searching for
// the "find" button as well.

void
XmlEditorDialogClass::FindAgainButton (wxCommandEvent & event)
{
#if 0
  wxMessageBox ("The search function is not implemented yet.",
  		"So very sorry!", wxCENTRE | wxOK | wxICON_INFORMATION);
#else // 0

  struct XmlStruct *Next, *RetVal;
  long from, to, n;
  wxString Value;
  int i;
  //int RetryCount = 0;
  
  //ItemText->SetFocus ();
	      
//Retry:
  //if (RetryCount++ > 1)
  //  return;  

  // Activate the FindAgain button.
  wxBitmapButton *FindagainbuttonPtr = GetFindagainbutton ();
  FindagainbuttonPtr->Enable (true);
  
  // What is the current item in the database?
  wxTreeItemId Selected = ItemTree->GetSelection ();
  struct XmlStruct *Current = FindXml (RootXml, Selected);
  if (Current == NULL)
    return;
    
  // Where is the cursor positioned?
  ItemText->GetSelection (&from, &to);
  if (from > to)
    to = from;
  if (to < 0)
    to = 0;
    
  // Get rid of everything prior to the current point.  Make lower
  // case so as to have case-insensitive search.  This may not 
  // correctly set the starting point of the search in Windows 
  // (because the selection pointers returned by GetSelection
  // may not correctly correspond to offsets in the string returned
  // by GetValue because of translation of "\r\n" to "\n"), but
  // I haven't a clue as to how to overcome that.
  Value = ItemText->GetValue ();
  Value = Value.Mid (to);
  Value = Value.MakeLower ();
  
  // Now, do we need to search the current item or not?  I don't 
  // think we should search the title or the hint, but we may
  // need to search the text.  We only need to do this if: 
  // a) we are searching text; and b) if it's the right document type.
  if (SearchText)
    {
      int DocTypeMatch;
      DocTypeMatch = SearchAll;
      if (!DocTypeMatch)
        {
	  for (i = 0; i < NUM_DOCUMENTS; i++)
	    if (SearchDocs[i] && Current->Documents[i] == DOCUMENT_YES)
	      {
	        DocTypeMatch = 1;
		break;
	      }
	}
      if (DocTypeMatch)
        {
	  n = Value.Find (SearchString.MakeLower ());
	  if (n != -1)	// Found, in current item!
	    {
	      to += n;
              
#if 0	      
              printf ("------------------------------------\n");
              printf ("FindAgainButton (1):  desire %ld %ld\n", to, to + SearchString.Length ());
#endif // 0	      
              
	      ItemText->SetInsertionPoint (to + SearchString.Length ());
	      //ItemText->SetFocus ();
	      //ItemText->ShowPosition (to);
	      ItemText->SetSelection (to, to + SearchString.Length ());

#if 0	      
	      {
		long nufrom, nuto;
		ItemText->GetSelection (&nufrom, &nuto);
		printf ("FindAgainButton (1):  actual %ld %ld\n", nufrom, nuto);
		//if (nufrom != to)
		  //{
		    //ItemText->SetInsertionPoint (to + SearchString.Length ());
		    //ItemText->SetSelection (to, to + SearchString.Length ());
		    //ItemText->SetSelection (to, to);
		    //goto Retry;
		  //}
	      }
#endif // 0	      
	      ItemText->SetFocus ();
	      return;
	    }
	}
    }
    
  // Well, we didn't find it in the current item, so let's check
  // out downstream items:  first children, then siblings, then
  // parents, grandparents, etc.
  // Seek the next item of this type.
  RetVal = ItemSearchByDocType3 (Current, 
				 CHILDREN,
				 (char *) SearchString.c_str (),
				 SearchDocs,
				 SearchAll,
				 SearchTitles,
				 SearchHints,
				 SearchText,
				 &from,
				 &to);
  if (RetVal == NULL)
    {
      RetVal = ItemSearchByDocType3 (Current, 
				     SIBLINGS,
				     (char *) SearchString.c_str (),
				     SearchDocs,
				     SearchAll,
				     SearchTitles,
				     SearchHints,
				     SearchText,
				     &from,
				     &to);
      if (RetVal == NULL)
	{
	  for (Next = Current; Next != NULL; Next = Next->Parent)
	    {
	      RetVal = ItemSearchByDocType3 (Next, 
					     PARENT,
					     (char *) SearchString.c_str (),
					     SearchDocs,
					     SearchAll,
					     SearchTitles,
					     SearchHints,
					     SearchText,
					     &from,
					     &to);
	      if (RetVal != NULL)
		break;
	    }
	}
    }

  // Open 'er up!
  if (RetVal != NULL)
    {
      wxTreeEvent DummyEvent;
      
      printf ("------------------------------------\n");
      printf ("FindAgainButton (2):  desire %ld %ld\n", from, to);

      for (Next = RetVal; Next != NULL; Next = Next->Parent)
	ItemTree->Expand (Next->Id);
      ItemTree->SelectItem (RetVal->Id);
      ClickTreeBypass = 1;
      ClickTree (DummyEvent);
      ItemText->SetInsertionPoint (to);
      //ItemText->SetFocus ();
      //ItemText->ShowPosition (from);
      ItemText->SetSelection (from, to);
#if 1
      {
        long nufrom, nuto;
	ItemText->GetSelection (&nufrom, &nuto);
        printf ("FindAgainButton (2):  actual %ld %ld\n", nufrom, nuto);
	//if (nufrom != from)
	  //{
	    //ItemText->SetInsertionPoint (to);
	    //ItemText->SetSelection (from, to);
	    //ItemText->SetSelection (from, from);
	    //goto Retry;
	  //}
      }
#endif // 0
      ItemText->SetFocus ();
    }

  
  		
#endif // 0
}


